@@ -6,24 +6,31 @@ Applies a Unfurl ensemble
 
 For each configuration, run it if required, then record the result
 """
-from __future__ import print_function
-
-from .job import runJob
-from .support import Status
-from . import __version__, versionTuple, initLogging, getHomeConfigPath, DefaultNames
-from . import init as initmod
-from .util import filterEnv, getPackageDigest
-from .localenv import LocalEnv, Project
-import click
-import sys
+import functools
+import getpass
+import json
+import logging
 import os
 import os.path
-import traceback
-import logging
-import functools
-import subprocess
+import re
 import shlex
-import json
+import shutil
+import subprocess
+import sys
+import traceback
+from pathlib import Path
+from typing import Optional
+
+import click
+
+from . import DefaultNames, __version__, get_home_config_path
+from . import init as initmod
+from . import logs, version_tuple
+from .job import start_job
+from .localenv import LocalEnv, Project
+from .logs import Levels
+from .support import Status
+from .util import filter_env, get_package_digest
 
 _latestJobs = []  # for testing
 _args = []  # for testing
@@ -74,11 +81,10 @@ def cli(
     ctx,
     verbose=0,
     quiet=False,
-    logfile=None,
     loglevel=None,
     tmp=None,
     version_check=None,
-    **kw
+    **kw,
 ):
     # ensure that ctx.obj exists and is a dict (in case `cli()` is called
     # by means other than the `if` block below
@@ -87,31 +93,13 @@ def cli(
 
     if tmp is not None:
         os.environ["UNFURL_TMPDIR"] = tmp
+    effective_log_level = detect_log_level(loglevel, quiet, verbose)
+    ctx.obj["verbose"] = detect_verbose_level(effective_log_level)
+    if kw["logfile"]:
+        logs.add_log_file(kw["logfile"])
+    logs.set_console_log_level(effective_log_level)
 
-    levels = [logging.INFO, 15, logging.DEBUG, 5, 5]
-    if quiet:
-        effectiveLogLevel = logging.CRITICAL
-    else:
-        # TRACE (5)
-        effectiveLogLevel = levels[min(verbose, 3)]
-
-    if loglevel:  # UNFURL_LOGGING overrides command line -v
-        effectiveLogLevel = dict(CRITICAL=50, ERROR=40, WARNING=30, INFO=20, DEBUG=10)[
-            loglevel.upper()
-        ]
-    # verbose: 0 == INFO, -1 == CRITICAL, >= 1 == DEBUG
-    if effectiveLogLevel == logging.CRITICAL:
-        verbose = -1
-    elif effectiveLogLevel >= logging.INFO:
-        verbose = 0
-    elif effectiveLogLevel == 15:
-        effectiveLogLevel = logging.DEBUG  # XXX add logging.VERBOSE and start using it
-        verbose = 1
-    else:  # must be DEBUG
-        verbose = 2
-    ctx.obj["verbose"] = verbose
-    initLogging(effectiveLogLevel, logfile)
-    if version_check and versionTuple() < versionTuple(version_check):
+    if version_check and version_tuple() < version_tuple(version_check):
         logging.warning(
             "current version %s older than expected version %s",
             __version__(True),
@@ -119,7 +107,35 @@ def cli(
         )
 
 
-jobControlOptions = option_group(
+def detect_log_level(loglevel: Optional[str], quiet: bool, verbose: int) -> Levels:
+    if quiet:
+        effective_log_level = Levels.CRITICAL
+    else:
+        if os.getenv("UNFURL_LOGGING"):
+            effective_log_level = Levels[os.getenv("UNFURL_LOGGING").upper()]
+        else:
+            levels = [Levels.INFO, Levels.VERBOSE, Levels.DEBUG, Levels.TRACE]
+            effective_log_level = levels[min(verbose, 3)]
+    if loglevel:
+        effective_log_level = Levels[loglevel.upper()]
+    return effective_log_level
+
+
+def detect_verbose_level(effective_log_level: Levels) -> int:
+    if effective_log_level is Levels.VERBOSE:
+        verbose = 1
+    elif effective_log_level is Levels.DEBUG:
+        verbose = 2
+    elif effective_log_level is Levels.TRACE:
+        verbose = 3
+    elif effective_log_level is Levels.CRITICAL:
+        verbose = -1
+    else:
+        verbose = 0
+    return verbose
+
+
+readonlyJobControlOptions = option_group(
     click.option(
         "--dryrun",
         default=False,
@@ -146,7 +162,16 @@ jobControlOptions = option_group(
         help="Set exit code to 1 if job status is not ok.",
     ),
 )
-
+jobControlOptions = option_group(
+    readonlyJobControlOptions,
+    click.option(
+        "--approve",
+        envvar="UNFURL_APPROVE",
+        default=False,
+        is_flag=True,
+        help="Don't prompt for approval to apply changes.",
+    ),
+)
 commonJobFilterOptions = option_group(
     click.option("--template", help="TOSCA template to target"),
     click.option("--instance", help="instance name to target"),
@@ -160,12 +185,18 @@ commonJobFilterOptions = option_group(
     ),
     click.option("--starttime", help="Set the start time of the job."),
     click.option(
-        "--destroyunmanaged",
+        "--force",
         default=False,
         is_flag=True,
-        help="include unmanaged instances for consideration when destroying",
+        help="(Re)run operation regardless of instance's status or state",
     ),
 )
+destroyUnmanagedOption = click.option(
+    "--destroyunmanaged",
+    default=False,
+    is_flag=True,
+    help="include unmanaged instances for consideration when destroying",
+)
 
 
 @cli.command(short_help="Run and record an ad-hoc command")
@@ -202,12 +233,12 @@ def run(ctx, instance="root", cmdline=None, **options):
     return _run(options.pop("ensemble"), options, ctx.info_name)
 
 
-def _getRuntime(options, ensemblePath):
+def _get_runtime(options, ensemblePath):
     runtime = options.get("runtime")
     localEnv = None
     if not runtime:
-        localEnv = LocalEnv(ensemblePath, options.get("home"))
-        runtime = localEnv.getRuntime()
+        localEnv = LocalEnv(ensemblePath, options.get("home"), can_be_empty=True)
+        runtime = localEnv.get_runtime()
     return runtime, localEnv
 
 
@@ -216,12 +247,12 @@ def _run(ensemble, options, workflow=None):
         options["workflow"] = workflow
 
     if not options.get("no_runtime"):
-        runtime, localEnv = _getRuntime(options, ensemble)
+        runtime, localEnv = _get_runtime(options, ensemble)
         if runtime and runtime != ".":
             if not localEnv:
                 localEnv = LocalEnv(ensemble, options.get("home"))
-            return _runRemote(runtime, options, localEnv)
-    return _runLocal(ensemble, options)
+            return _run_remote(runtime, options, localEnv)
+    return _run_local(ensemble, options)
 
 
 def _venv(runtime, env):
@@ -234,12 +265,17 @@ def _venv(runtime, env):
     return env
 
 
-def _remoteCmd(runtime, cmdLine, localEnv):
-    context = localEnv.getContext()
-    kind, sep, rest = runtime.partition(":")
-    if context.get("environment"):
+def _remote_cmd(runtime_, cmd_line, local_env):
+    context = local_env.get_context()
+    kind, sep, rest = runtime_.partition(":")
+    envvar_filter = context.get("variables") or {}
+    for name in ["UNFURL_APPROVE", "UNFURL_LOGGING"]:
+        if name in os.environ:
+            envvar_filter[name] = os.environ[name]
+
+    if envvar_filter:
         addOnly = kind == "docker"
-        env = filterEnv(localEnv.mapValue(context["environment"]), addOnly=addOnly)
+        env = filter_env(local_env.map_value(envvar_filter), addOnly=addOnly)
     else:
         env = None
 
@@ -255,27 +291,107 @@ def _remoteCmd(runtime, cmdLine, localEnv):
                 "--version-check",
                 __version__(True),
             ]
-            + cmdLine,
+            + cmd_line,
             False,
         )
-    # elif docker: docker $container -it run $cmdline
+    elif kind == "docker":
+        cmd = DockerCmd(runtime_, env or {}).build()
+        return env, cmd + cmd_line, False
     else:
         # treat as shell command
-        cmd = shlex.split(runtime)
+        cmd = shlex.split(runtime_)
         return (
             env,
-            cmd + ["--no-runtime", "--version-check", __version__(True)] + cmdLine,
+            cmd + ["--no-runtime", "--version-check", __version__(True)] + cmd_line,
             True,
         )
 
 
-def _runRemote(runtime, options, localEnv):
+class DockerCmd:
+    """Builds command for docker runtime"""
+
+    def __init__(self, specifier_string: str, env_vars: dict) -> None:
+        self.env_vars = env_vars
+        # if running from a development branch use "latest" otherwise the image for this release
+        tag = "latest" if len(version_tuple()) > 3 else __version__()
+        self.image = self.parse_image(specifier_string, tag)
+        self.docker_args = self.parse_docker_args(specifier_string)
+
+    @staticmethod
+    def parse_image(specifier_string, version):
+        strings = specifier_string.split(maxsplit=1)
+        image = strings[0]
+        image = re.sub(r"^docker:*", "", image)  # remove prefix
+        if image:
+            if ":" in image:
+                return image
+            else:
+                return f"{image}:{version}"
+        return f"onecommons/unfurl:{version}"
+
+    @staticmethod
+    def parse_docker_args(specifier_string):
+        strings = specifier_string.split(maxsplit=1)
+        if len(strings) == 2:
+            return strings[1].split()
+        return []
+
+    def build(self) -> list:
+        """Prepare command which will be run as subprocess"""
+
+        cmd = [
+            "docker",
+            "run",
+            "--rm",
+            "-w",
+            "/data",
+            "-u",
+            f"{os.getuid()}:{os.getgid()}",
+        ]
+        if sys.stdout.isatty():
+            cmd.append("-it")
+        cmd.extend(self.env_vars_to_args())
+        cmd.extend(self.default_volumes())
+        cmd.extend(self.docker_args)
+        cmd.extend(
+            [self.image, "unfurl", "--no-runtime", "--version-check", __version__(True)]
+        )
+        return cmd
+
+    def env_vars_to_args(self) -> list:
+        user = getpass.getuser()
+        args = [
+            "-e",
+            f"HOME=/home/{user}",
+            "-e",
+            f"USER={user}",
+        ]
+        for k, v in self.env_vars.items():
+            args.extend(["-e", f"{k}={v}"])
+        return args
+
+    @staticmethod
+    def default_volumes() -> list:
+        """Volumes for docker command"""
+        user = getpass.getuser()
+        return [
+            "-v",
+            f"{Path.cwd()}:/data",
+            "-v",
+            f"{Path.home()}:/home/{user}",
+            "-v",
+            "/var/run/docker.sock:/var/run/docker.sock",
+        ]
+
+
+def _run_remote(runtime, options, localEnv):
     logger = logging.getLogger("unfurl")
     logger.debug('running command remotely on "%s"', runtime)
     cmdLine = _args or sys.argv[1:]
-    if _args:
-        print("TESTING: running remote with _args %s" % _args)
-    env, remote, shell = _remoteCmd(runtime, cmdLine, localEnv)
+    if _args:  # set by test driver to override command line
+        print(f"TESTING: running remote with _args {_args}")
+    env, remote, shell = _remote_cmd(runtime, cmdLine, localEnv)
+    logger.debug("executing remote command: %s", remote)
     rv = subprocess.call(remote, env=env, shell=shell)
     if options.get("standalone_mode") is False:
         return rv
@@ -283,37 +399,91 @@ def _runRemote(runtime, options, localEnv):
         sys.exit(rv)
 
 
-def _runLocal(ensemble, options):
-    logger = logging.getLogger("unfurl")
-    job = runJob(ensemble, options)
-    _latestJobs.append(job)
-    if not job:
-        click.echo("Unable to create job")
-    elif job.unexpectedAbort:
-        click.echo("Job unexpected aborted")
-        if options.get("verbose", 0) > 0:
-            raise job.unexpectedAbort
+def _print_summary(job, options):
+    jsonSummary = {}
+    summary = options.get("output")
+    if summary == "text" and not job.jobOptions.planOnly:
+        click.echo(job.summary())
+    elif summary == "json":
+        if job.jobOptions.planOnly:
+            jsonSummary = job._json_plan_summary()
+        else:
+            jsonSummary = job.json_summary()
+
+    query = options.get("query")
+    if query:
+        result = job.run_query(query, options.get("trace"))
+        if summary == "json":
+            jsonSummary["query"] = query
+            jsonSummary["result"] = result
+        else:
+            click.echo("query: " + query)
+            click.echo(result)
+    if jsonSummary is not None:
+        click.echo(json.dumps(jsonSummary, indent=2))
+
+
+def yesno(prompt):
+    click.echo(prompt + " [yN] ", nl=False)
+    c = click.getchar()
+    click.echo(c)
+    if c == "y" or c == "Y":
+        return True
     else:
-        jsonSummary = {}
-        summary = options.get("output")
-        logger.debug(job.summary())
-        if summary == "text":
-            click.echo(job.summary())
-        elif summary == "json":
-            jsonSummary = job.jsonSummary()
+        return False
 
-        query = options.get("query")
-        if query:
-            result = job.runQuery(query, options.get("trace"))
-            if summary == "json":
-                jsonSummary["query"] = query
-                jsonSummary["result"] = result
+
+def _stop_logging(job, options, verbose, tmplogfile):
+    if options["logfile"]:
+        if verbose > -1:
+            click.echo("Done, full log appended to " + options["logfile"])
+    else:
+        if job and job.log_path:
+            log_path = job.log_path
+            dir = os.path.dirname(log_path)
+            if not os.path.exists(dir):
+                os.makedirs(dir)
+            try:
+                os.rename(tmplogfile, log_path)
+            except OSError:
+                # handle [Errno 18] Invalid cross-device link
+                shutil.copy(tmplogfile, log_path)
+        else:
+            log_path = tmplogfile
+        if verbose > -1:
+            click.echo("Done, full log written to " + log_path)
+
+
+def _run_local(ensemble, options):
+    verbose = options.get("verbose", 0)
+    tmplogfile = None
+    if not options["logfile"]:
+        tmplogfile = logs.get_tmplog_path()
+        logs.add_log_file(tmplogfile)
+
+    job, rendered, proceed = start_job(ensemble, options)
+    _latestJobs.append(job)  # testing only
+    if job:
+        declined = False
+        if not job.unexpectedAbort and not job.planOnly and proceed:
+            if options.get("approve") or yesno("proceed with job?"):
+                job.run(rendered)
             else:
-                click.echo("query: " + query)
-                click.echo(result)
-        if jsonSummary is not None:
-            click.echo(json.dumps(jsonSummary, indent=2))
+                declined = True
+        if job.unexpectedAbort:
+            click.echo("Job unexpected aborted")
+            if verbose > 0:
+                raise job.unexpectedAbort
+        elif not declined:
+            _print_summary(job, options)
+    else:
+        click.echo("Unable to create job")
 
+    _stop_logging(job, options, verbose, tmplogfile)
+    return _exit(job, options)
+
+
+def _exit(job, options):
     if not job or (
         "jobexitcode" in options
         and options["jobexitcode"] != "never"
@@ -327,34 +497,34 @@ def _runLocal(ensemble, options):
         return 0
 
 
-# XXX update help text sans "configurations"
-deployFilterOptions = option_group(
+checkFilterOptions = option_group(
     click.option(
-        "--add", default=True, is_flag=True, help="run newly added configurations"
-    ),
-    click.option(
-        "--update",
-        default=True,
+        "--skip-new",
+        default=False,
         is_flag=True,
-        help="run configurations that whose spec has changed but don't require a major version change",
+        help="Don't create instance for new templates.",
     ),
+    click.option(
+        "--change-detection",
+        default="evaluate",
+        type=click.Choice(["skip", "spec", "evaluate"]),
+        help="How to detect configuration changes to existing instances. (Default: evaluate)",
+    ),
+)
+
+deployFilterOptions = option_group(
+    checkFilterOptions,
     click.option(
         "--repair",
-        type=click.Choice(["error", "degraded", "missing", "none"]),
+        type=click.Choice(["error", "degraded", "none"]),
         default="error",
-        help="re-run configurations that are in an error or degraded state",
+        help="Re-run operations on instances that are in an error or degraded state. (Default: error)",
     ),
     click.option(
         "--upgrade",
         default=False,
         is_flag=True,
-        help="run configurations with major version changes or whose spec has changed",
-    ),
-    click.option(
-        "--force",
-        default=False,
-        is_flag=True,
-        help="(re)run operation regardless of instance's status or state",
+        help="Apply major versions changes.",
     ),
     click.option(
         "--prune",
@@ -362,6 +532,7 @@ deployFilterOptions = option_group(
         is_flag=True,
         help="destroy instances that are no longer used",
     ),
+    destroyUnmanagedOption,
     click.option(
         "--check",
         default=False,
@@ -389,12 +560,14 @@ def deploy(ctx, ensemble=None, **options):
 @click.pass_context
 @click.argument("ensemble", default="", type=click.Path(exists=False))
 @commonJobFilterOptions
-@jobControlOptions
+@checkFilterOptions
+@readonlyJobControlOptions
 def check(ctx, ensemble=None, **options):
     """
     Check and update the status of the ensemble's instances
     """
     options.update(ctx.obj)
+    options["approve"] = True
     return _run(ensemble, options, ctx.info_name)
 
 
@@ -416,6 +589,7 @@ def discover(ctx, ensemble=None, **options):
 @click.argument("ensemble", default="", type=click.Path(exists=False))
 @commonJobFilterOptions
 @jobControlOptions
+@destroyUnmanagedOption
 def undeploy(ctx, ensemble=None, **options):
     """
     Destroy what was deployed.
@@ -444,7 +618,7 @@ def stop(ctx, ensemble=None, **options):
 @deployFilterOptions
 @click.option("--workflow", default="deploy", help="plan workflow (default: deploy)")
 def plan(ctx, ensemble=None, **options):
-    "Print the given deployment plan"
+    """Print the given deployment plan"""
     options.update(ctx.obj)
     options["planOnly"] = True
     # XXX show status and task to run including preview of generated templates, cmds to run etc.
@@ -454,8 +628,12 @@ def plan(ctx, ensemble=None, **options):
 @cli.command(short_help="Create a new unfurl project or ensemble")
 @click.pass_context
 @click.argument("projectdir", default="", type=click.Path(exists=False))
+@click.argument("ensemble_name", default="")
 @click.option(
-    "--mono", default=False, is_flag=True, help="Create one repository for the project."
+    "--mono",
+    default=False,
+    is_flag=True,
+    help="Don't create a separate ensemble repository.",
 )
 @click.option(
     "--existing",
@@ -474,12 +652,31 @@ def plan(ctx, ensemble=None, **options):
 )
 @click.option(
     "--template",
-    type=click.Path(exists=True),
+    type=click.Path(exists=False),
     help="Absolute path to a directory of project templates.",
 )
-def init(ctx, projectdir, **options):
+@click.option(
+    "--create-environment",
+    is_flag=True,
+    default=False,
+    help="Create (if missing) a environment with the same name and set this as the default repository for ensembles that use this environment.",
+)
+@click.option(
+    "--use-environment",
+    default=None,
+    help="Associate the given environment with this ensemble.",
+)
+@click.option(
+    "--shared-repository",
+    default=None,
+    type=click.Path(exists=True),
+    help="Create the ensemble in an repository outside the project.",
+)
+def init(ctx, projectdir, ensemble_name=None, **options):
+    """
+    Create a new project or, if [project_dir] exists or is inside a project, create a new ensemble.
+    If [ensemble_name] is omitted, use a default name.
     """
-    Create a new project or, if [project_dir] exists or is inside a project, create a new ensemble"""
     options.update(ctx.obj)
     if not projectdir:
         # if adding a project to an existing repository use '.unfurl' as the default name
@@ -488,14 +685,15 @@ def init(ctx, projectdir, **options):
         else:  # otherwise use the current directory
             projectdir = "."
 
-    projectPath = Project.findPath(projectdir)
+    projectPath = Project.find_path(projectdir)
     if projectPath:
         # dest is already in a project, so create a new ensemble in it instead of a new project
         projectPath = os.path.dirname(projectPath)  # strip out unfurl.yaml
         # if projectPath is deeper than projectDir (happens if it is .unfurl) set projectDir to that
         if len(os.path.abspath(projectPath)) > len(os.path.abspath(projectdir)):
             projectdir = projectPath
-        message = initmod.clone(projectPath, projectdir, **options)
+        # this will clone the default ensemble if it exists or use ensemble-template
+        message = initmod.clone(projectPath, projectdir, ensemble_name, **options)
         click.echo(message)
         return
     if os.path.exists(projectdir):
@@ -510,12 +708,12 @@ def init(ctx, projectdir, **options):
                 'Can not create project in "' + projectdir + '": folder is not empty'
             )
 
-    homePath, projectPath, repo = initmod.createProject(
-        os.path.abspath(projectdir), **options
+    homePath, projectPath, repo = initmod.create_project(
+        os.path.abspath(projectdir), ensemble_name, **options
     )
     if homePath:
-        click.echo("unfurl home created at %s" % homePath)
-    click.echo("New Unfurl project created at %s" % projectPath)
+        click.echo(f"unfurl home created at {homePath}")
+    click.echo(f"New Unfurl project created at {projectPath}")
 
 
 # XXX add --upgrade option
@@ -541,17 +739,17 @@ def home(ctx, init=False, render=False, replace=False, **options):
     options.update(ctx.obj)
     if not render and not init:
         # just display the current home location
-        click.echo(getHomeConfigPath(options.get("home")))
+        click.echo(get_home_config_path(options.get("home")))
         return
 
-    homePath = initmod.createHome(render=render, replace=replace, **options)
+    homePath = initmod.create_home(render=render, replace=replace, **options)
     action = "rendered" if render else "created"
     if homePath:
-        click.echo("unfurl home %s at %s" % (action, homePath))
+        click.echo(f"unfurl home {action} at {homePath}")
     else:
-        currentHome = getHomeConfigPath(options.get("home"))
+        currentHome = get_home_config_path(options.get("home"))
         if currentHome:
-            click.echo("Can't %s home, it already exists at %s" % (action, currentHome))
+            click.echo(f"Can't {action} home, it already exists at {currentHome}")
         else:
             click.echo("Error: home path is empty")
 
@@ -565,32 +763,32 @@ def home(ctx, init=False, render=False, replace=False, **options):
     default=".",
 )
 def runtime(ctx, project_folder, init=False, **options):
-    """If no options are set, display the runtime currently used by the project. To create a new runtime in the project root
-    use --init and the global --runtime option.
+    """If no options are set, display the runtime currently used by the project. To create
+    a new runtime in the project root use --init and the global --runtime option.
     """
     options.update(ctx.obj)
-    projectPath = os.path.abspath(project_folder)
+    project_path = os.path.abspath(project_folder)
     if not init:
         # just display the current runtime
         try:
-            runtime, localEnv = _getRuntime(options, projectPath)
+            runtime_, local_env = _get_runtime(options, project_folder)
         except:
             # if the current path isn't a folder
             if project_folder == ".":
-                runtime, localEnv = _getRuntime(
-                    options, getHomeConfigPath(options.get("home"))
+                runtime_, local_env = _get_runtime(
+                    options, get_home_config_path(options.get("home"))
                 )
             else:
                 raise
-        click.echo(runtime)
+        click.echo(f"\nCurrent runtime: {runtime_}")
         return
 
-    runtime = options.get("runtime") or "venv:"
-    error = initmod.initEngine(projectPath, runtime)
+    runtime_ = options.get("runtime") or "venv:"
+    error = initmod.init_engine(project_path, runtime_)
     if not error:
-        click.echo('Created runtime "%s" in "%s"' % (runtime, projectPath))
+        click.echo(f'Created runtime "{runtime_}" in "{project_path}"')
     else:
-        click.echo('Failed to create runtime "%s": %s' % (runtime, error))
+        click.echo(f'Failed to create runtime "{runtime_}": {error}')
 
 
 @cli.command(short_help="Clone a project, ensemble or service template")
@@ -612,6 +810,15 @@ def runtime(ctx, project_folder, init=False, **options):
     is_flag=True,
     help="Add project to nearest existing repository.",
 )
+@click.option(
+    "--empty", default=False, is_flag=True, help="Don't create a default ensemble."
+)
+# XXX
+# @click.option(
+#     "--use-environment",
+#     default=None,
+#     help="Associate the given environment with the ensemble.",
+# )
 def clone(ctx, source, dest, **options):
     """Create a new ensemble or project from a service template or an existing ensemble or project.
 
@@ -632,21 +839,38 @@ def clone(ctx, source, dest, **options):
 )
 @click.pass_context
 @click.option(
-    "--dir", default=".", type=click.Path(exists=True), help="path to spec repository"
+    "--dir",
+    default=".",
+    type=click.Path(exists=True),
+    help='Path to project or ensemble (default: ".")',
 )
 @click.argument("gitargs", nargs=-1)
 def git(ctx, gitargs, dir="."):
     """
-    unfurl git --dir=/path/to/start [gitoptions] [gitcmd] [gitcmdoptions]: Runs command on each project repository."""
-    localEnv = LocalEnv(dir, ctx.obj.get("home"))
-    repos = localEnv.getRepos()
+    unfurl git [git command] [git command arguments]: Run the given git command on each project repository."""
+    localEnv = LocalEnv(dir, ctx.obj.get("home"), can_be_empty=True)
+    if localEnv.manifestPath:
+        repos = {
+            os.path.relpath(repo.working_dir, os.path.abspath(dir)): repo.repo
+            for repo in localEnv.get_manifest().repositories.values()
+            if repo.repo
+        }
+    elif localEnv.project.project_repoview.repo:
+        repo = localEnv.project.project_repoview.repo
+        repos = {os.path.relpath(repo.working_dir, os.path.abspath(dir)): repo}
+    else:
+        repos = {}
+
     status = 0
     if not repos:
         click.echo("Can't run git command, no repositories found")
-    for repo in repos:
-        repoPath = os.path.relpath(repo.workingDir, os.path.abspath(dir))
-        click.echo("*** Running 'git %s' in './%s'" % (" ".join(gitargs), repoPath))
-        _status, stdout, stderr = repo.runCmd(gitargs)
+    # sort by working_dir for determinism
+    for working_dir in sorted(repos):
+        repo = repos[working_dir]
+        if working_dir != ".":
+            working_dir = "./" + working_dir
+        click.echo(f"*** Running 'git {' '.join(gitargs)}' in '{working_dir}'")
+        _status, stdout, stderr = repo.run_cmd(gitargs)
         click.echo(stdout + "\n")
         if stderr.strip():
             click.echo(stderr + "\n", color="red")
@@ -656,14 +880,13 @@ def git(ctx, gitargs, dir="."):
     return status
 
 
-def get_commit_message(manifest):
-    statuses = manifest.getRepoStatuses(True)
+def get_commit_message(committer):
+    statuses = committer.get_repo_status(True)
     if not statuses:
         click.echo("Nothing to commit!")
         return None
-    defaultMsg = manifest.getDefaultCommitMessage()
+    defaultMsg = committer.get_default_commit_message()
     MARKER = "# Everything below is ignored\n"
-    statuses = "".join(statuses)
     statusMsg = "# " + "\n# ".join(statuses.rstrip().splitlines())
     message = click.edit(defaultMsg + "\n\n" + MARKER + statusMsg)
     if message is not None:
@@ -675,7 +898,7 @@ def get_commit_message(manifest):
 
 @cli.command()
 @click.pass_context
-@click.option("--ensemble", default="", type=click.Path(exists=False))
+@click.argument("project_or_ensemble_path", default=".", type=click.Path(exists=True))
 @click.option("-m", "--message", help="commit message to use")
 @click.option(
     "--no-edit",
@@ -687,46 +910,66 @@ def get_commit_message(manifest):
     "--skip-add",
     default=False,
     is_flag=True,
-    help="Don't add files for committing (if set, must first manually add)",
+    help="Don't add files for committing (user must add using git)",
 )
-def commit(ctx, message, ensemble, skip_add, no_edit, **options):
-    "Commit any outstanding changes to this ensemble."
+def commit(ctx, project_or_ensemble_path, message, skip_add, no_edit, **options):
+    """Commit any outstanding changes to the given project or ensemble."""
     options.update(ctx.obj)
-    localEnv = LocalEnv(ensemble, options.get("home"))
-    manifest = localEnv.getManifest()
+    localEnv = LocalEnv(
+        project_or_ensemble_path, options.get("home"), can_be_empty=True
+    )
+    if localEnv.manifestPath:
+        committer = localEnv.get_manifest()
+    else:
+        committer = localEnv.project.project_repoview
+
     if not skip_add:
-        manifest.addAll()
+        committer.add_all()
     if not message:
         if no_edit:
-            message = manifest.getDefaultCommitMessage()
+            message = committer.get_default_commit_message()
         else:
-            message = get_commit_message(manifest)
+            message = get_commit_message(committer)
             if not message:
                 return  # aborted
-    committed = manifest.commit(message, False)
-    click.echo("committed to %s repositories" % committed)
+    committed = committer.commit(message, False)
+    click.echo(f"committed to {committed} repositories")
 
 
-@cli.command(short_help="Show the git status for paths relevant to this ensemble.")
+@cli.command()
 @click.pass_context
-@click.option("--ensemble", default="", type=click.Path(exists=False))
+@click.argument("project_or_ensemble_path", default=".", type=click.Path(exists=True))
 @click.option(
     "--dirty",
     default=False,
     is_flag=True,
     help="Only show repositories with uncommitted changes",
 )
-def git_status(ctx, ensemble, dirty, **options):
-    "Show the git status for repository paths that are relevant to this ensemble."
+def git_status(ctx, project_or_ensemble_path, dirty, **options):
+    "Show the git status for paths relevant to the given project or ensemble."
     options.update(ctx.obj)
-    localEnv = LocalEnv(ensemble, options.get("home"))
-    manifest = localEnv.getManifest()
-    statuses = manifest.getRepoStatuses(dirty)
+    localEnv = LocalEnv(
+        project_or_ensemble_path, options.get("home"), can_be_empty=True
+    )
+    if localEnv.manifestPath:
+        committer = localEnv.get_manifest()
+    else:
+        committer = localEnv.project.project_repoview
+    statuses = committer.get_repo_status(dirty)
     if not statuses:
         click.echo("No status to display.")
     else:
-        for status in statuses:
-            click.echo(status)
+        click.echo(statuses)
+
+
+@cli.command(short_help="Show the git status for paths relevant to this ensemble.")
+@click.pass_context
+@click.argument("ensemble", default=".", type=click.Path(exists=True))
+def status(ctx, ensemble, **options):
+    """Show the status of deployed resources in the given ensemble."""
+    options.update(ctx.obj)
+    localEnv = LocalEnv(ensemble, options.get("home"))
+    click.echo(localEnv.get_manifest().status_summary())
 
 
 @cli.command()
@@ -744,23 +987,23 @@ def git_status(ctx, ensemble, dirty, **options):
     help="Also print the version installed in the current runtime.",
 )
 def version(ctx, semver=False, remote=False, **options):
-    "Print the current version"
+    """Print the current version"""
     options.update(ctx.obj)
     if semver:
         click.echo(__version__())
     else:
-        click.echo("unfurl version %s (%s)" % (__version__(True), getPackageDigest()))
+        click.echo(f"unfurl version {__version__(True)} ({get_package_digest()})")
 
     if remote and not options.get("no_runtime"):
         click.echo("Remote:")
-        _run(getHomeConfigPath(options.get("home")), options)
+        _run(get_home_config_path(options.get("home")), options)
 
 
 @cli.command()
 @click.pass_context
 @click.argument("cmd", nargs=1, default="")
 def help(ctx, cmd=""):
-    "Get help on a command"
+    """Get help on a command"""
     if not cmd:
         click.echo(cli.get_help(ctx.parent), color=ctx.color)
         return
@@ -770,7 +1013,7 @@ def help(ctx, cmd=""):
         ctx.info_name = cmd  # hack
         click.echo(command.get_help(ctx), color=ctx.color)
     else:
-        raise click.UsageError("no help found for unknown command '%s'" % cmd, ctx=ctx)
+        raise click.UsageError(f"no help found for unknown command '{cmd}'", ctx=ctx)
 
 
 def main():
@@ -784,7 +1027,7 @@ def main():
     except click.ClickException as e:
         if obj.get("verbose", 0) > 0:
             traceback.print_exc(file=sys.stderr)
-        click.secho("Error: %s" % e.format_message(), fg="red", err=True)
+        click.secho(f"Error: {e.format_message()}", fg="red", err=True)
         sys.exit(e.exit_code)
     except Exception as err:
         if obj.get("verbose", 0) > 0:
@@ -796,12 +1039,12 @@ def main():
 
 def vaultclient():
     try:
-        localEnv = LocalEnv(".")
+        localEnv = LocalEnv(".", can_be_empty=True)
     except Exception as err:
         click.echo(str(err), err=True)
         return 1
-
-    print(localEnv.getVaultPassword() or "")
+    # XXX check for --vault-id and pass to get_vault_password
+    print(localEnv.get_vault_password() or "")
     return 0
 
 
