@@ -1,11 +1,11 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
-import collections
+from collections.abc import Mapping
 import os.path
 import hashlib
 import json
 from ruamel.yaml.comments import CommentedMap
-from .tosca import ToscaSpec, TOSCA_VERSION, Artifact
+from .tosca import ToscaSpec, TOSCA_VERSION, ArtifactSpec
 
 from .support import ResourceChanges, AttributeManager, Status, Priority, NodeState
 from .runtime import (
@@ -13,9 +13,11 @@ from .runtime import (
     NodeInstance,
     CapabilityInstance,
     RelationshipInstance,
+    ArtifactInstance,
 )
-from .util import UnfurlError, toEnum, sensitive_str, getBaseDir
-from .repo import RevisionManager, splitGitUrl, RepoView
+from .util import UnfurlError, to_enum, sensitive_str, get_base_dir
+from .repo import RevisionManager, split_git_url, RepoView
+from .merge import patch_dict
 from .yamlloader import YamlConfig, yaml, ImportResolver
 from .job import ConfigChange
 import toscaparser.imports
@@ -38,10 +40,10 @@ class Manifest(AttributeManager):
     rootResource = None
 
     def __init__(self, path, localEnv=None):
-        super(Manifest, self).__init__(yaml)
+        super().__init__(yaml)
         self.localEnv = localEnv
         self.path = path
-        self.repo = self._findRepo()
+        self.repo = self._find_repo()
         self.currentCommitId = self.repo and self.repo.revision
         self.revisions = RevisionManager(self)
         self.changeSets = None
@@ -49,7 +51,7 @@ class Manifest(AttributeManager):
         self.specDigest = None
         self.repositories = {}
 
-    def _setSpec(self, config):
+    def _set_spec(self, config, more_spec=None):
         """
         Set the TOSCA service template.
         """
@@ -57,19 +59,19 @@ class Manifest(AttributeManager):
             name: repo.repository.tpl for name, repo in self.repositories.items()
         }
         spec = config.get("spec", {})
-        self.tosca = self._loadSpec(spec, self.path, repositories)
-        self.specDigest = self.getSpecDigest(spec)
+        self.tosca = self._load_spec(spec, self.path, repositories, more_spec)
+        self.specDigest = self.get_spec_digest(spec)
 
-    def _findRepo(self):
+    def _find_repo(self):
         # check if this path exists in the repo
         repo = self.localEnv and self.localEnv.instanceRepo
         if repo:
-            path = repo.findPath(self.path)[0]
+            path = repo.find_path(self.path)[0]
             if path and (path, 0) in repo.repo.index.entries:
                 return repo
         return None
 
-    def _loadSpec(self, spec, path, repositories):
+    def _load_spec(self, spec, path, repositories, more_spec):
         if "service_template" in spec:
             toscaDef = spec["service_template"] or {}
         elif "tosca" in spec:  # backward compat
@@ -85,16 +87,19 @@ class Manifest(AttributeManager):
         # make sure this is present
         toscaDef["tosca_definitions_version"] = TOSCA_VERSION
 
+        if more_spec:
+            patch_dict(toscaDef, more_spec, True)
+
         if not isinstance(toscaDef, CommentedMap):
             toscaDef = CommentedMap(toscaDef.items())
-        if getattr(toscaDef, "baseDir", None) and (
-            not path or toscaDef.baseDir != os.path.dirname(path)
+        if getattr(toscaDef, "base_dir", None) and (
+            not path or toscaDef.base_dir != os.path.dirname(path)
         ):
             # note: we only recorded the baseDir not the name of the included file
-            path = toscaDef.baseDir
-        return ToscaSpec(toscaDef, spec, path, self.getImportResolver())
+            path = toscaDef.base_dir
+        return ToscaSpec(toscaDef, spec, path, self.get_import_resolver(expand=True))
 
-    def getSpecDigest(self, spec):
+    def get_spec_digest(self, spec):
         m = hashlib.sha1()  # use same digest function as git
         t = self.tosca.template
         for tpl in [spec, t.topology_template.custom_defs, t.nested_tosca_tpls]:
@@ -109,10 +114,10 @@ class Manifest(AttributeManager):
         if rootResource:
             rootResource.attributeManager = self
 
-    def getRootResource(self):
+    def get_root_resource(self):
         return self.rootResource
 
-    def getBaseDir(self):
+    def get_base_dir(self):
         return "."
 
     @property
@@ -123,52 +128,52 @@ class Manifest(AttributeManager):
             and self.rootResource.templar._loader
         )
 
-    def saveJob(self, job):
+    def save_job(self, job):
         pass
 
-    def loadTemplate(self, name, lastChange=None):
+    def load_template(self, name, lastChange=None):
         if lastChange:
             try:
-                return self.revisions.getRevision(lastChange).tosca.getTemplate(name)
+                return self.revisions.get_revision(lastChange).tosca.get_template(name)
             except:
                 return None
         else:
-            return self.tosca.getTemplate(name)
+            return self.tosca.get_template(name)
 
     #  load instances
     #    create a resource with the given template
     #  or generate a template setting interface with the referenced implementations
 
     @staticmethod
-    def loadStatus(status, instance=None):
+    def load_status(status, instance=None):
         if not instance:
             instance = OperationalInstance()
         if not status:
             return instance
 
-        instance._priority = toEnum(Priority, status.get("priority"))
+        instance._priority = to_enum(Priority, status.get("priority"))
         instance._lastStateChange = status.get("lastStateChange")
         instance._lastConfigChange = status.get("lastConfigChange")
 
         readyState = status.get("readyState")
-        if not isinstance(readyState, collections.Mapping):
-            instance._localStatus = toEnum(Status, readyState)
+        if not isinstance(readyState, Mapping):
+            instance._localStatus = to_enum(Status, readyState)
         else:
-            instance._localStatus = toEnum(Status, readyState.get("local"))
-            instance._state = toEnum(NodeState, readyState.get("state"))
+            instance._localStatus = to_enum(Status, readyState.get("local"))
+            instance._state = to_enum(NodeState, readyState.get("state"))
 
         return instance
 
     @staticmethod
-    def loadResourceChanges(changes):
+    def load_resource_changes(changes):
         resourceChanges = ResourceChanges()
         if changes:
             for k, change in changes.items():
                 status = change.pop(".status", None)
                 if isinstance(status, dict):
-                    status = Manifest.loadStatus(status).localStatus
+                    status = Manifest.load_status(status).local_status
                 else:
-                    status = toEnum(Status, status)
+                    status = to_enum(Status, status)
                 resourceChanges[k] = [
                     status,
                     change.pop(".added", None),
@@ -176,14 +181,14 @@ class Manifest(AttributeManager):
                 ]
         return resourceChanges
 
-    def loadConfigChange(self, changeSet):
+    def load_config_change(self, changeSet):
         """
         Reconstruct the Configuration that was applied in the past
         """
         from .configurator import Dependency
 
         configChange = ConfigChange()
-        Manifest.loadStatus(changeSet, configChange)
+        Manifest.load_status(changeSet, configChange)
         configChange.changeId = changeSet.get("changeId", 0)
         configChange.previousId = changeSet.get("previousId")
         configChange.target = changeSet.get("target")
@@ -209,7 +214,7 @@ class Manifest(AttributeManager):
             )
 
         if "changes" in changeSet:
-            configChange.resourceChanges = self.loadResourceChanges(
+            configChange.resourceChanges = self.load_resource_changes(
                 changeSet["changes"]
             )
 
@@ -242,72 +247,82 @@ class Manifest(AttributeManager):
     #         postConditions=spec.get("postConditions"),
     #     )
 
-    def createNodeInstance(self, rname, resourceSpec, parent=None):
+    def _create_requirement(self, key, val):
+        # parent will be the capability, should have already been created
+        capabilityId = val.get("capability")
+        if not capabilityId:
+            nodeId = val.get("node")
+            if not nodeId:
+                raise UnfurlError(f"requirement is missing capability {key}")
+        capability = capabilityId and self.get_root_resource().query(capabilityId)
+        if not capability or not isinstance(capability, CapabilityInstance):
+            raise UnfurlError(f"can not find capability {capabilityId}")
+        if capability._relationships is None:
+            capability._relationships = []
+        return self._create_entity_instance(RelationshipInstance, key, val, capability)
+
+    def create_node_instance(self, rname, resourceSpec, parent=None):
         # if parent property is set it overrides the parent argument
         pname = resourceSpec.get("parent")
         if pname:
-            parent = self.getRootResource().findResource(pname)
+            parent = self.get_root_resource().find_resource(pname)
             if parent is None:
-                raise UnfurlError("can not find parent instance %s" % pname)
+                raise UnfurlError(f"can not find parent instance {pname}")
 
-        resource = self._createEntityInstance(NodeInstance, rname, resourceSpec, parent)
+        resource = self._create_entity_instance(
+            NodeInstance, rname, resourceSpec, parent
+        )
         if resourceSpec.get("capabilities"):
             for key, val in resourceSpec["capabilities"].items():
-                self._createEntityInstance(CapabilityInstance, key, val, resource)
+                self._create_entity_instance(CapabilityInstance, key, val, resource)
 
         if resourceSpec.get("requirements"):
-            for key, val in resourceSpec["requirements"].items():
-                # parent will be the capability, should have already been created
-                capabilityId = val.get("capability")
-                if not capabilityId:
-                    raise UnfurlError("requirement is missing capability %s" % key)
-                capability = capabilityId and self.getRootResource().query(capabilityId)
-                if not capability or not isinstance(capability, CapabilityInstance):
-                    raise UnfurlError("can not find capability %s" % capabilityId)
-                if capability._relationships is None:
-                    capability._relationships = []
-                requirement = self._createEntityInstance(
-                    RelationshipInstance, key, val, capability
-                )
+            for req in resourceSpec["requirements"]:
+                key, val = next(iter(req.items()))
+                requirement = self._create_requirement(key, val)
                 requirement.source = resource
                 resource.requirements.append(requirement)
 
+        if resourceSpec.get("artifacts"):
+            for key, val in resourceSpec["artifacts"].items():
+                self._create_entity_instance(ArtifactInstance, key, val, resource)
+
         for key, val in resourceSpec.get("instances", {}).items():
-            self.createNodeInstance(key, val, resource)
+            self.create_node_instance(key, val, resource)
 
         return resource
 
-    def _getLastChange(self, operational):
-        if not operational.lastConfigChange:
+    def _get_last_change(self, operational):
+        if not operational.last_config_change:
             return None
         if not self.changeSets:  # XXX load changesets if None
             return None
-        jobId = ConfigChange.getJobId(operational.lastConfigChange)
+        jobId = ConfigChange.get_job_id(operational.last_config_change)
         return self.changeSets.get(jobId)
 
-    def _createEntityInstance(self, ctor, name, status, parent):
-        operational = self.loadStatus(status)
+    def _create_entity_instance(self, ctor, name, status, parent):
+        operational = self.load_status(status)
         templateName = status.get("template", name)
 
         imported = None
         importName = status.get("imported")
         if importName is not None:
-            imported = self.imports.findImport(importName)
+            imported = self.imports.find_import(importName)
             if not imported:
-                raise UnfurlError("missing import %s" % importName)
+                raise UnfurlError(f"missing import {importName}")
             if imported.shadow:
-                raise UnfurlError("already imported %s" % importName)
+                raise UnfurlError(f"already imported {importName}")
             template = imported.template
         else:
-            template = self.loadTemplate(templateName)
+            template = self.load_template(templateName)
 
         if template is None:
             # not defined in the current model any more, try to retrieve the old version
-            if operational.lastConfigChange:
-                changerecord = self._getLastChange(operational)
-                template = self.loadTemplate(templateName, changerecord)
+            if operational.last_config_change:
+                changerecord = self._get_last_change(operational)
+                template = self.load_template(templateName, changerecord)
         if template is None:
-            raise UnfurlError("missing resource template %s" % templateName)
+            raise UnfurlError(f"missing resource template {templateName}")
         # logger.debug("creating instance for template %s: %s", templateName, template)
 
         # omit keys that match <<REDACTED>> so can we use the computed property
@@ -320,21 +335,21 @@ class Manifest(AttributeManager):
         if "created" in status:
             instance.created = status["created"]
         if imported:
-            self.imports.setShadow(importName, instance)
+            self.imports.set_shadow(importName, instance)
         return instance
 
-    def statusSummary(self):
+    def status_summary(self):
         def summary(instance, indent):
-            output.append(
-                "%s%s %s %s %s"
-                % (
-                    " " * indent,
-                    instance,
-                    instance.status,
-                    instance.state,
-                    instance.created,
-                )
-            )
+            status = "" if instance.status is None else instance.status.name
+            state = instance.state and instance.state.name or ""
+            if instance.created:
+                if isinstance(instance.created, bool):
+                    created = "managed"
+                else:
+                    created = f"created by {instance.created}"
+            else:
+                created = ""
+            output.append(f"{' ' * indent}{instance} {status} {state} {created}")
             indent += 4
             for child in instance.instances:
                 summary(child, indent)
@@ -343,13 +358,13 @@ class Manifest(AttributeManager):
         summary(self.rootResource, 0)
         return "\n".join(output)
 
-    def findPathInRepos(self, path, importLoader=None):  # currently unused
+    def find_path_in_repos(self, path, importLoader=None):  # currently unused
         """
         Check if the file path is inside a folder that is managed by a repository.
         If the revision is pinned and doesn't match the repo, it might be bare
         """
         if self.localEnv:
-            return self.localEnv.findPathInRepos(path, importLoader)
+            return self.localEnv.find_path_in_repos(path, importLoader)
         elif self.repo:
             repo = self.repo
             filePath = repo.findRepoPath(path)
@@ -358,50 +373,61 @@ class Manifest(AttributeManager):
         return None, None, None, None
 
     # NOTE: all the methods below may be called during config parse time via loadYamlInclude()
-    def findRepoFromGitUrl(self, path, isFile, importLoader):
-        repoURL, filePath, revision = splitGitUrl(path)
+    def find_repo_from_git_url(self, path, isFile, importLoader):
+        repoURL, filePath, revision = split_git_url(path)
         if not repoURL:
-            raise UnfurlError("invalid git URL %s" % path)
+            raise UnfurlError(f"invalid git URL {path}")
         assert self.localEnv
-        basePath = getBaseDir(importLoader.path)  # checks if dir or not
-        repo, revision, bare = self.localEnv.findOrCreateWorkingDir(
+        basePath = get_base_dir(importLoader.path)  # checks if dir or not
+        repo, revision, bare = self.localEnv.find_or_create_working_dir(
             repoURL, revision, basePath
         )
         return repo, filePath, revision, bare
 
-    def addRepository(self, repo, toscaRepository, file_name):
+    def add_repository(self, repo, toscaRepository, file_name):
         repository = self.repositories.get(toscaRepository.name)
         if repository:
+            # already exist, make sure it's the same repo
             repo = repo or repository.repo
             if (
                 repository.repo and repository.repo != repo
             ) or repository.repository.tpl != toscaRepository.tpl:
                 raise UnfurlError(
-                    'Repository "%s" already defined' % toscaRepository.name
+                    f'Repository "{toscaRepository.name}" already defined'
                 )
-        self.repositories[repository.name] = RepoView(toscaRepository, repo, file_name)
+        self.repositories[toscaRepository.name] = RepoView(
+            toscaRepository, repo, file_name
+        )
         return repository
 
-    def updateRepositories(self, config, inlineRepositories=None):
-        repositories = self._getRepositories(config)
+    def update_repositories(self, config, inlineRepositories=None, resolver=None):
+        if not resolver:
+            resolver = self.get_import_resolver(self)
+        repositories = self._get_repositories(config)
         for name, tpl in repositories.items():
             if name not in self.repositories:
-                toscaRepository = Repository(name, tpl)
+                toscaRepository = resolver.get_repository(name, tpl)
                 self.repositories[name] = RepoView(toscaRepository, None)
         if inlineRepositories:
-            for repository in inlineRepositories:
+            for name, repository in inlineRepositories.items():
+                if name in self.repositories:
+                    logger.verbose(
+                        'skipping inline repository definition for "%s", it was previously defined',
+                        name,
+                    )
+                    continue
                 if isinstance(repository, dict):
-                    repository = Repository(repository.pop("name"), repository)
-                self.addRepository(None, repository, "")
-        return self._setRepositories()
+                    repository = Repository(name, repository)
+                self.add_repository(None, repository, "")
+        return self._set_repositories()
 
     @staticmethod
-    def _getRepositories(tpl):
+    def _get_repositories(tpl):
         return ((tpl.get("spec") or {}).get("service_template") or {}).get(
             "repositories"
         ) or {}
 
-    def _setRepositories(self):
+    def _set_repositories(self):
         repositories = self.repositories
         if "unfurl" not in repositories:
             # add a repository that points to this package
@@ -413,10 +439,10 @@ class Manifest(AttributeManager):
             repositories["unfurl"] = repository
         if "self" not in repositories:
             # this is called too early to use self.getBaseDir()
-            path = getBaseDir(self.path) if self.path else "."
+            path = get_base_dir(self.path) if self.path else "."
             if self.repo:
-                url = self.repo.getUrlWithPath(path)
-                path = os.path.relpath(path, self.repo.workingDir)
+                url = self.repo.get_url_with_path(path)
+                path = os.path.relpath(path, self.repo.working_dir)
             else:
                 url = "file:" + path
             repository = RepoView(
@@ -425,36 +451,25 @@ class Manifest(AttributeManager):
                 path,
             )
             repositories["self"] = repository
+
+        inProject = False
+        if self.localEnv and self.localEnv.project:
+            if self.localEnv.project is self.localEnv.homeProject:
+                inProject = self.localEnv.project.projectRoot in self.path
+            else:
+                inProject = True
+        if inProject and "project" not in repositories:
+            repositories["project"] = self.localEnv.project.project_repoview
+
         if "spec" not in repositories:
             # if not found assume it points the project root or self if not in a project
-            inProject = False
-            if self.localEnv and self.localEnv.project:
-                if self.localEnv.project is self.localEnv.homeProject:
-                    inProject = self.localEnv.project.projectRoot in self.path
-                else:
-                    inProject = True
-
             if inProject:
-                repo = self.localEnv.project.projectRepo
-                path = self.localEnv.project.projectRoot
-                if repo:
-                    assert path
-                    url = repo.getUrlWithPath(path)
-                    path = os.path.relpath(path, repo.workingDir)
-                else:
-                    url = "file:" + path
-
-                repository = RepoView(
-                    Repository("spec", dict(url=url)),
-                    self.localEnv.project.projectRepo,
-                    path,
-                )
-                repositories["spec"] = repository
+                repositories["spec"] = self.localEnv.project.project_repoview
             else:
                 repositories["spec"] = repositories["self"]
         return {name: repo.repository.tpl for name, repo in self.repositories.items()}
 
-    def loadYamlInclude(
+    def load_yaml_include(
         self,
         yamlConfig,
         templatePath,
@@ -478,7 +493,8 @@ class Manifest(AttributeManager):
                 reponame = repo.setdefault("name", os.path.basename(path))
                 # replace spec with just its name
                 artifactTpl["repository"] = reponame
-                inlineRepository = repo
+
+                inlineRepository = {reponame: repo}
         else:
             artifactTpl = dict(file=templatePath)
 
@@ -487,44 +503,42 @@ class Manifest(AttributeManager):
                 reponame = artifactTpl["repository"]
                 if reponame in ["spec", "self", "unfurl"]:  # builtin
                     return True
-                repositories = self._getRepositories(expanded)
+                repositories = self._get_repositories(expanded)
                 return reponame in repositories
             return True
 
-        artifact = Artifact(artifactTpl, path=baseDir)
-
-        extraRepos = []
-        if inlineRepository:
-            extraRepos.append(inlineRepository)
+        artifact = ArtifactSpec(artifactTpl, path=baseDir)
 
         tpl = CommentedMap()
-        tpl["repositories"] = self.updateRepositories(
-            expanded or yamlConfig.config, extraRepos
+
+        resolver = self.get_import_resolver(warnWhenNotFound)
+        tpl["repositories"] = self.update_repositories(
+            expanded or yamlConfig.config, inlineRepository, resolver
         )
         loader = toscaparser.imports.ImportsLoader(
             None,
-            artifact.baseDir,
+            artifact.base_dir,
             tpl=tpl,
-            resolver=self.getImportResolver(warnWhenNotFound),
+            resolver=resolver,
         )
-        return loader._load_import_template(None, artifact.asImportSpec())
+        return loader._load_import_template(None, artifact.as_import_spec())
 
-    def getImportResolver(self, ignoreFileNotFound=False):
-        return ImportResolver(self, ignoreFileNotFound)
+    def get_import_resolver(self, ignoreFileNotFound=False, expand=False):
+        return ImportResolver(self, ignoreFileNotFound, expand)
 
 
 class SnapShotManifest(Manifest):
     def __init__(self, manifest, commitId):
-        super(SnapShotManifest, self).__init__(manifest.path, self.localEnv)
+        super().__init__(manifest.path, self.localEnv)
         self.commitId = commitId
         oldManifest = manifest.repo.show(manifest.path, commitId)
         self.repo = manifest.repo
         self.localEnv = manifest.localEnv
         self.manifest = YamlConfig(
-            oldManifest, manifest.path, loadHook=self.loadYamlInclude
+            oldManifest, manifest.path, loadHook=self.load_yaml_include
         )
-        self.updateRepositories(oldManifest)
+        self.update_repositories(oldManifest)
         expanded = self.manifest.expanded
         # just needs the spec, not root resource
-        self._setSpec(expanded)
+        self._set_spec(expanded)
         self._ready(None)
