@@ -3,7 +3,8 @@
 import itertools
 import re
 import six
-from collections import Mapping, MutableSequence, namedtuple, Sequence
+from collections import namedtuple
+from collections.abc import Mapping, MutableSequence, Sequence
 
 from ruamel.yaml.comments import CommentedMap, CommentedBase
 
@@ -11,13 +12,27 @@ from .util import UnfurlError
 
 
 def _mapCtor(self):
-    if hasattr(self, "baseDir"):
-        return makeMapWithBase(self, self.baseDir)
+    if hasattr(self, "base_dir"):
+        return make_map_with_base(self, self.base_dir)
     return CommentedMap
 
 
 CommentedMap.mapCtor = property(_mapCtor)
 
+# we can't subclass CommentedMap so we need to monkey patch
+__base__deepcopy__ = CommentedMap.__deepcopy__
+
+
+def __deepcopy__(self, memo):
+    cp = __base__deepcopy__(self, memo)
+    if hasattr(self, "base_dir"):
+        setattr(cp, "base_dir", self.base_dir)
+    return cp
+
+
+if CommentedMap.__deepcopy__ is not __deepcopy__:
+    CommentedMap.__deepcopy__ = __deepcopy__
+
 
 def copy(src):
     cls = getattr(src, "mapCtor", src.__class__)
@@ -26,7 +41,7 @@ def copy(src):
     return cls(src)
 
 
-def makeMapWithBase(doc, baseDir):
+def make_map_with_base(doc, baseDir):
     loadTemplate = getattr(doc, "loadTemplate", None)
     _anchorCache = getattr(doc, "_anchorCache", None)
 
@@ -35,7 +50,7 @@ def makeMapWithBase(doc, baseDir):
             map = CommentedMap(args[0].items(), **kws)
         else:
             map = CommentedMap(*args, **kws)
-        map.baseDir = baseDir
+        map.base_dir = baseDir
         if loadTemplate:
             map.loadTemplate = loadTemplate
         if _anchorCache is not None:
@@ -47,10 +62,17 @@ def makeMapWithBase(doc, baseDir):
 
 # XXX?? because json keys are strings allow number keys to merge with lists
 # other values besides delete not supported because current code can leave those keys in final result
-mergeStrategyKey = "+%"  # values: delete
+mergeStrategyKey = "+%"  # supported values: "whiteout", "nullout"
 
 # b is the merge patch, a is original dict
-def mergeDicts(b, a, cls=None, replaceKeys=None, defaultStrategy="merge"):
+def merge_dicts(
+    b,
+    a,
+    cls=None,
+    replaceKeys=None,
+    defaultStrategy="merge",
+    listStrategy="append_unique",
+):
     """
     Returns a new dict (or cls) that recursively merges b into a.
     b is base, a overrides.
@@ -67,8 +89,9 @@ def mergeDicts(b, a, cls=None, replaceKeys=None, defaultStrategy="merge"):
             childStrategy = "replace"
         else:
             childStrategy = "merge"
-        if isinstance(val, Mapping):
-            strategy = val.get(mergeStrategyKey)
+        # note: for merging treat None as an empty map
+        if isinstance(val, Mapping) or val is None:
+            strategy = val and val.get(mergeStrategyKey)
             if key in b:
                 bval = b[key]
                 if isinstance(bval, Mapping):
@@ -78,11 +101,12 @@ def mergeDicts(b, a, cls=None, replaceKeys=None, defaultStrategy="merge"):
                     if strategy == "merge":
                         if not val:  # empty map, treat as missing key
                             continue
-                        cp[key] = mergeDicts(
+                        cp[key] = merge_dicts(
                             bval,
                             val,
                             defaultStrategy=childStrategy,
                             replaceKeys=replaceKeys,
+                            listStrategy=listStrategy,
                         )
                         continue
                     if strategy == "error":
@@ -90,12 +114,14 @@ def mergeDicts(b, a, cls=None, replaceKeys=None, defaultStrategy="merge"):
                             "merging %s is not allowed, +%%: error was set" % key
                         )
                 # otherwise we ignore bval because key is already in a
-            if strategy == "delete":
+            if strategy == "whiteout":
                 skip.append(key)
                 continue
+            if strategy == "nullout":
+                val = None
         elif isinstance(val, MutableSequence) and key in b:
             bval = b[key]
-            if isinstance(bval, MutableSequence):
+            if isinstance(bval, MutableSequence) and listStrategy == "append_unique":
                 # XXX allow more strategies beyond append
                 #     if appendlists == 'all' or key in appendlists:
                 cp[key] = bval + [item for item in val if item not in bval]
@@ -123,7 +149,9 @@ def mergeDicts(b, a, cls=None, replaceKeys=None, defaultStrategy="merge"):
     return cp
 
 
-def _cacheAnchors(_anchorCache, obj):
+def _cache_anchors(_anchorCache, obj):
+    if not hasattr(obj, "yaml_anchor"):
+        return
     anchor = obj.yaml_anchor()
     if anchor and anchor.value:
         _anchorCache[anchor.value] = obj
@@ -133,10 +161,10 @@ def _cacheAnchors(_anchorCache, obj):
 
     for value in obj.values() if isinstance(obj, dict) else obj:
         if isinstance(value, CommentedBase):
-            _cacheAnchors(_anchorCache, value)
+            _cache_anchors(_anchorCache, value)
 
 
-def findAnchor(doc, anchorName):
+def find_anchor(doc, anchorName):
     if not isinstance(doc, CommentedMap):
         return None
 
@@ -144,20 +172,20 @@ def findAnchor(doc, anchorName):
     if _anchorCache is None:
         _anchorCache = {}
         # recursively find anchors
-        _cacheAnchors(_anchorCache, doc)
+        _cache_anchors(_anchorCache, doc)
         doc._anchorCache = _anchorCache
 
     return _anchorCache.get(anchorName)
 
 
-def _jsonPointerUnescape(s):
+def _json_pointer_unescape(s):
     return s.replace("~1", "/").replace("~0", "~")
 
 
 _RE_INVALID_JSONPOINTER_ESCAPE = re.compile("(~[^01]|~$)")
 
 
-def _jsonPointerValidate(pointer):
+def _json_pointer_validate(pointer):
     invalid_escape = _RE_INVALID_JSONPOINTER_ESCAPE.search(pointer)
     if invalid_escape:
         raise UnfurlError(
@@ -168,18 +196,18 @@ def _jsonPointerValidate(pointer):
     return None
 
 
-def getTemplate(doc, key, value, path, cls, includes=None):
+def get_template(doc, key, value, path, cls, includes=None):
     template = doc
     templatePath = None
     if key.include:
         value, template, baseDir = doc.loadTemplate(value, key.maybe, doc)
         if template is None:  # include wasn't not found and key.maybe
             return None
-        cls = makeMapWithBase(doc, baseDir)
+        cls = make_map_with_base(doc, baseDir)
         # fileKey = key._replace(include=None)
         # template = getTemplate(template, fileKey, "raw", (), cls)
     else:
-        result = _findTemplate(doc, key, path, cls, not key.maybe)
+        result = _find_template(doc, key, path, cls, not key.maybe)
         if result is None:
             return doc
         template, templatePath = result
@@ -199,28 +227,27 @@ def getTemplate(doc, key, value, path, cls, includes=None):
                 )
                 if len(prefix) == len(templatePath):
                     raise UnfurlError(
-                        'recursive include "%s" in "%s" when including %s'
-                        % (templatePath, path, key.key)
+                        f'recursive include "{templatePath}" in "{path}" when including {key.key}'
                     )
             if includes is None:
                 includes = CommentedMap()
-            template = expandDict(doc, path, includes, template, cls=cls)
+            template = expand_dict(doc, path, includes, template, cls=cls)
     finally:
         if key.include:
             doc.loadTemplate(baseDir)  # pop baseDir
     return template
 
 
-def _findTemplate(doc, key, path, cls, fail):
+def _find_template(doc, key, path, cls, fail):
     template = doc
     templatePath = None
     if key.anchor:
-        template = findAnchor(doc, key.anchor)
+        template = find_anchor(doc, key.anchor)
         if template is None:
             if not fail:
                 return None
             else:
-                raise UnfurlError("could not find anchor '%s'" % key.anchor)
+                raise UnfurlError(f"could not find anchor '{key.anchor}'")
         # XXX we don't know the path to the anchor so we can't support relative paths also
     elif key.relative:
         stop = None
@@ -230,12 +257,12 @@ def _findTemplate(doc, key, path, cls, fail):
             templatePath = list(path[:stop])
         else:
             templatePath = templatePath[:stop]
-        template = lookupPath(doc, templatePath, cls)
+        template = lookup_path(doc, templatePath, cls)
         if template is None:
             if not fail:
                 return None
             else:
-                raise UnfurlError("could relative path '%s'" % ("." * key.relative))
+                raise UnfurlError(f"could relative path '{('.' * key.relative)}'")
     for index, segment in enumerate(key.pointer):
         if isinstance(template, Sequence):
             # Array indexes should be turned into integers
@@ -257,16 +284,16 @@ def _findTemplate(doc, key, path, cls, fail):
     return template, templatePath
 
 
-def hasTemplate(doc, key, value, path, cls):
+def has_template(doc, key, value, path, cls):
     if key.include:
         loadTemplate = getattr(doc, "loadTemplate", None)
         if not loadTemplate:
             return False
         return doc.loadTemplate(value, key.maybe, doc, True)
-    return _findTemplate(doc, key, path, cls, False) is not None
+    return _find_template(doc, key, path, cls, False) is not None
 
 
-class _MissingInclude(object):
+class _MissingInclude:
     def __init__(self, key, value):
         self.key = key
         self.value = value
@@ -281,7 +308,7 @@ RE_FIRST = re.compile(r"([?]?)(include\d*)?([*]\S+)?([.]+$)?")
 MergeKey = namedtuple("MergeKey", "key, maybe, include, anchor, relative, pointer")
 
 
-def parseMergeKey(key):
+def parse_merge_key(key):
     """
     +[maybe]?[include]?[anchor]?[relative]?[jsonpointer]?
 
@@ -291,8 +318,8 @@ def parseMergeKey(key):
     """
     original = key
     key = key[1:]
-    _jsonPointerValidate(key)
-    parts = [_jsonPointerUnescape(part) for part in key.split("/")]
+    _json_pointer_validate(key)
+    parts = [_json_pointer_unescape(part) for part in key.split("/")]
     first = parts.pop(0)
     maybe, include, anchor, relative = RE_FIRST.match(first).groups()
     if anchor:
@@ -311,7 +338,7 @@ def parseMergeKey(key):
     return MergeKey(original, not not maybe, include, anchor, relative, tuple(parts))
 
 
-def expandDict(doc, path, includes, current, cls=dict):
+def expand_dict(doc, path, includes, current, cls=dict):
     """
     Return a copy of `doc` that expands include directives.
     Include directives look like "+path/to/value"
@@ -332,32 +359,34 @@ def expandDict(doc, path, includes, current, cls=dict):
             if key == mergeStrategyKey:
                 cp[key] = value
                 continue
-            mergeKey = parseMergeKey(key)
+            mergeKey = parse_merge_key(key)
             if not mergeKey:
                 cp[key] = value
                 continue
-            foundTemplate = hasTemplate(doc, mergeKey, value, path, cls)
+            foundTemplate = has_template(doc, mergeKey, value, path, cls)
             if not foundTemplate:
                 includes.setdefault(path, []).append(_MissingInclude(mergeKey, value))
                 cp[key] = value
                 continue
             includes.setdefault(path, []).append((mergeKey, value))
-            template = getTemplate(doc, mergeKey, value, path, cls, includes)
+            template = get_template(doc, mergeKey, value, path, cls, includes)
             if isinstance(template, Mapping):
                 templates.append(template)
             elif mergeKey.include and template is None:
                 continue  # include path not found
             else:
                 if len(current) > 1:  # XXX include merge directive keys in count
-                    raise UnfurlError("can not merge non-map value %s" % template)
+                    raise UnfurlError(
+                        f"can not merge non-map value of type {type(template)}: {template}"
+                    )
                 else:
                     return template  # current dict is replaced with a value
         # elif key.startswith("q+"):
         #    cp[key[2:]] = value
         elif isinstance(value, Mapping):
-            cp[key] = expandDict(doc, path + (key,), includes, value, cls)
+            cp[key] = expand_dict(doc, path + (key,), includes, value, cls)
         elif isinstance(value, list):
-            cp[key] = list(expandList(doc, path + (key,), includes, value, cls))
+            cp[key] = list(expand_list(doc, path + (key,), includes, value, cls))
         else:
             cp[key] = value
 
@@ -366,7 +395,7 @@ def expandDict(doc, path, includes, current, cls=dict):
         templates.append(cp)
         while templates:
             cls = getattr(templates[0], "mapCtor", cls)
-            accum = mergeDicts(accum, templates.pop(0), cls)
+            accum = merge_dicts(accum, templates.pop(0), cls)
         return accum
     else:
         return cp
@@ -374,53 +403,53 @@ def expandDict(doc, path, includes, current, cls=dict):
     # return includes, reduce(lambda accum, next: mergeDicts(accum, next, cls), templates, {}), cp
 
 
-def _findMissingIncludes(includes):
+def _find_missing_includes(includes):
     for x in includes.values():
         for i in x:
             if isinstance(i, _MissingInclude):
                 yield i
 
 
-def _deleteDeletedKeys(expanded):
+def _delete_deleted_keys(expanded):
     for key, value in expanded.items():
         if isinstance(value, Mapping):
-            if value.get(mergeStrategyKey) == "delete":
+            if value.get(mergeStrategyKey) == "whiteout":
                 del expanded[key]
             else:
-                _deleteDeletedKeys(value)
+                _delete_deleted_keys(value)
 
 
-def expandDoc(doc, current=None, cls=dict):
+def expand_doc(doc, current=None, cls=dict):
     includes = CommentedMap()
     if current is None:
         current = doc
     if not isinstance(doc, Mapping) or not isinstance(current, Mapping):
-        raise UnfurlError("top level element %s is not a dict" % doc)
-    expanded = expandDict(doc, (), includes, current, cls)
+        raise UnfurlError(f"top level element {doc} is not a dict")
+    expanded = expand_dict(doc, (), includes, current, cls)
     if hasattr(doc, "_anchorCache"):
         expanded._anchorCache = doc._anchorCache
     last = 0
     while True:
-        missing = list(_findMissingIncludes(includes))
+        missing = list(_find_missing_includes(includes))
         if len(missing) == 0:
             # remove any stray keys with delete merge directive
-            _deleteDeletedKeys(expanded)
+            _delete_deleted_keys(expanded)
             return includes, expanded
         if len(missing) == last:  # no progress
-            raise UnfurlError("missing includes: %s" % missing)
+            raise UnfurlError(f"missing includes: {missing}")
         last = len(missing)
         includes = CommentedMap()
-        expanded = expandDict(expanded, (), includes, current, cls)
+        expanded = expand_dict(expanded, (), includes, current, cls)
         if hasattr(doc, "_anchorCache"):
             expanded._anchorCache = doc._anchorCache
 
 
-def expandList(doc, path, includes, value, cls=dict):
+def expand_list(doc, path, includes, value, cls=dict):
     for i, item in enumerate(value):
         if isinstance(item, Mapping):
-            if item.get(mergeStrategyKey) == "delete":
+            if item.get(mergeStrategyKey) == "whiteout":
                 continue
-            newitem = expandDict(doc, path + (i,), includes, item, cls)
+            newitem = expand_dict(doc, path + (i,), includes, item, cls)
             if isinstance(newitem, MutableSequence):
                 for i in newitem:
                     yield i
@@ -430,22 +459,28 @@ def expandList(doc, path, includes, value, cls=dict):
             yield item
 
 
-def diffDicts(old, new, cls=dict):
+def diff_dicts(old, new, cls=dict):
     """
-    return a dict where old + diff = new
+    given old, new return diff where merge_dicts(old, diff) == new
     """
     diff = cls()
     # start with old to preserve original order
-    for key, val in old.items():
+    for key, oldval in old.items():
         if key in new:
             newval = new[key]
-            if val != newval:
-                if isinstance(val, Mapping) and isinstance(newval, Mapping):
-                    diff[key] = diffDicts(val, newval, cls)
+            if oldval != newval:
+                if isinstance(oldval, Mapping):
+                    if isinstance(newval, Mapping):
+                        diff[key] = diff_dicts(oldval, newval, cls)
+                    elif newval is None:
+                        # dicts merge with None so add a nullout directive to preserve the None
+                        diff[key] = cls((("+%", "nullout"),))
+                    else:  # new non-dict val replaces old dict
+                        diff[key] = newval
                 else:
                     diff[key] = newval
-        else:
-            diff[key] = {"+%": "delete"}
+        else:  # not in new, so add a whiteout directive to delete this key
+            diff[key] = cls((("+%", "whiteout"),))
 
     for key in new:
         if key not in old:
@@ -454,9 +489,11 @@ def diffDicts(old, new, cls=dict):
 
 
 # XXX rename function, confusing name
-def patchDict(old, new, preserve=False):
+def patch_dict(old, new, preserve=False):
     """
-    Transform old into new while preserving old as much as possible.
+    Transform old into new based on object equality while preserving as much of old object as possible.
+
+    If ``preserve`` is True ``new`` will be merged in without removing ``old`` items.
     """
     # start with old to preserve original order
     for key, val in list(old.items()):
@@ -464,15 +501,19 @@ def patchDict(old, new, preserve=False):
             newval = new[key]
             if val != newval:
                 if isinstance(val, Mapping) and isinstance(newval, Mapping):
-                    old[key] = patchDict(val, newval, preserve)
+                    old[key] = patch_dict(val, newval, preserve)
                 elif isinstance(val, MutableSequence) and isinstance(
                     newval, MutableSequence
                 ):
-                    # preserve old item in list if they are equal to the new item
-                    old[key] = [
-                        (val[val.index(item)] if item in val else item)
-                        for item in newval
-                    ]
+                    if preserve:
+                        # add new items that don't appear in old list
+                        old[key] = val + [item for item in newval if item not in val]
+                    else:
+                        # preserve old item in list if they are equal to the new item
+                        old[key] = [
+                            (val[val.index(item)] if item in val else item)
+                            for item in newval
+                        ]
                 else:
                     old[key] = newval
         elif not preserve:
@@ -485,7 +526,7 @@ def patchDict(old, new, preserve=False):
     return old
 
 
-def intersectDict(old, new, cls=dict):
+def intersect_dict(old, new, cls=dict):
     """
     remove keys from old that don't match new
     """
@@ -495,7 +536,7 @@ def intersectDict(old, new, cls=dict):
             newval = new[key]
             if val != newval:
                 if isinstance(val, Mapping) and isinstance(newval, Mapping):
-                    old[key] = intersectDict(val, newval, cls)
+                    old[key] = intersect_dict(val, newval, cls)
                 else:
                     del old[key]
         else:
@@ -504,7 +545,7 @@ def intersectDict(old, new, cls=dict):
     return old
 
 
-def lookupPath(doc, path, cls=dict):
+def lookup_path(doc, path, cls=dict):
     template = doc
     for segment in path:
         if isinstance(template, Sequence):
@@ -519,19 +560,19 @@ def lookupPath(doc, path, cls=dict):
     return template
 
 
-def replacePath(doc, key, value, cls=dict):
+def replace_path(doc, key, value, cls=dict):
     path = key[:-1]
     last = key[-1]
-    ref = lookupPath(doc, path, cls)
+    ref = lookup_path(doc, path, cls)
     ref[last] = value
 
 
-def addTemplate(changedDoc, path, mergeKey, template, cls):
+def add_template(changedDoc, path, mergeKey, template, cls):
     # if includeKey.anchor: #???
     if mergeKey.relative:
         if mergeKey.relative > 1:
             path = path[: (mergeKey.relative - 1) * -1]
-        current = lookupPath(changedDoc, path, cls)
+        current = lookup_path(changedDoc, path, cls)
     else:
         current = changedDoc
 
@@ -543,7 +584,7 @@ def addTemplate(changedDoc, path, mergeKey, template, cls):
     current[last] = template
 
 
-def restoreIncludes(includes, originalDoc, changedDoc, cls=dict):
+def restore_includes(includes, originalDoc, changedDoc, cls=dict):
     """
     Modifies changedDoc with to use the includes found in originalDoc
     """
@@ -551,9 +592,9 @@ def restoreIncludes(includes, originalDoc, changedDoc, cls=dict):
     # resolve the include
     # if the include doesn't exist in the current doc, re-add it
     # create a diff between the current object and the merged includes
-    expandedOriginalIncludes, expandedOriginalDoc = expandDoc(originalDoc, cls=cls)
+    expandedOriginalIncludes, expandedOriginalDoc = expand_doc(originalDoc, cls=cls)
     for key, value in includes.items():
-        ref = lookupPath(changedDoc, key, cls)
+        ref = lookup_path(changedDoc, key, cls)
         if ref is None:
             # inclusion point no longer exists
             continue
@@ -563,18 +604,18 @@ def restoreIncludes(includes, originalDoc, changedDoc, cls=dict):
             if includeKey.include:
                 ref = None
                 continue
-            stillHasTemplate = hasTemplate(
+            stillHasTemplate = has_template(
                 changedDoc, includeKey, includeValue, key, cls
             )
             if stillHasTemplate:
-                template = getTemplate(changedDoc, includeKey, includeValue, key, cls)
+                template = get_template(changedDoc, includeKey, includeValue, key, cls)
             else:
-                if hasTemplate(originalDoc, includeKey, includeValue, key, cls):
-                    template = getTemplate(
+                if has_template(originalDoc, includeKey, includeValue, key, cls):
+                    template = get_template(
                         originalDoc, includeKey, includeValue, key, cls
                     )
                 else:
-                    template = getTemplate(
+                    template = get_template(
                         expandedOriginalDoc, includeKey, includeValue, key, cls
                     )
 
@@ -582,7 +623,7 @@ def restoreIncludes(includes, originalDoc, changedDoc, cls=dict):
                 # XXX3 if isinstance(ref, list) lists not yet implemented
                 if ref == template:
                     # ref still resolves to the template's value so replace it with the include
-                    replacePath(changedDoc, key, {includeKey.key: includeValue}, cls)
+                    replace_path(changedDoc, key, {includeKey.key: includeValue}, cls)
                 # ref isn't a map anymore so can't include a template
                 break
 
@@ -590,19 +631,21 @@ def restoreIncludes(includes, originalDoc, changedDoc, cls=dict):
                 # ref no longer includes that template or we don't want to save it
                 continue
             else:
-                mergedIncludes = mergeDicts(mergedIncludes, template, cls)
+                mergedIncludes = merge_dicts(mergedIncludes, template, cls)
                 ref[includeKey.key] = includeValue
 
             if not stillHasTemplate:
                 if includeValue != "raw":
-                    if hasTemplate(originalDoc, includeKey, includeValue, key, cls):
-                        template = getTemplate(originalDoc, includeKey, "raw", key, cls)
+                    if has_template(originalDoc, includeKey, includeValue, key, cls):
+                        template = get_template(
+                            originalDoc, includeKey, "raw", key, cls
+                        )
                     else:
-                        template = getTemplate(
+                        template = get_template(
                             expandedOriginalDoc, includeKey, "raw", key, cls
                         )
-                addTemplate(changedDoc, key, includeKey, template, cls)
+                add_template(changedDoc, key, includeKey, template, cls)
 
         if isinstance(ref, Mapping):
-            diff = diffDicts(mergedIncludes, ref, cls)
-            replacePath(changedDoc, key, diff, cls)
+            diff = diff_dicts(mergedIncludes, ref, cls)
+            replace_path(changedDoc, key, diff, cls)
