@@ -8,15 +8,21 @@ Each instance have a status; attributes that describe its state; and a TOSCA tem
 which describes its capabilities, relationships and available interfaces for configuring and interacting with it.
 """
 import six
-import collections
+from collections.abc import Mapping
 
 from ansible.parsing.dataloader import DataLoader
 
-from .util import UnfurlError, loadClass, toEnum, makeTempDir, ChainMap
+from .util import UnfurlError, load_class, to_enum, make_temp_dir, ChainMap
 from .result import ResourceRef, ChangeAware
 
 from .support import AttributeManager, Defaults, Status, Priority, NodeState, Templar
-from .tosca import CapabilitySpec, RelationshipSpec, NodeSpec, TopologySpec
+from .tosca import (
+    CapabilitySpec,
+    RelationshipSpec,
+    NodeSpec,
+    TopologySpec,
+    ArtifactSpec,
+)
 
 import logging
 
@@ -39,21 +45,24 @@ class Operational(ChangeAware):
         return Defaults.shouldRun
 
     @property
-    def localStatus(self):
+    def local_status(self):
         return Status.unknown
 
     @property
     def state(self):
         return NodeState.initial
 
-    def getOperationalDependencies(self):
+    def get_operational_dependencies(self):
+        """
+        Return an iterator of `Operational` object that this instance depends on to be operational.
+        """
         return ()
 
-    def getOperationalDependents(self):
+    def get_operational_dependents(self):
         return ()
 
     @property
-    def manualOverideStatus(self):
+    def manual_overide_status(self):
         return None
 
     # derived properties:
@@ -77,19 +86,19 @@ class Operational(ChangeAware):
     def status(self):
         """
         Return the effective status, considering first the local readyState and
-        then the aggregate status' of its dependencies (see `aggregateStatus()`
-        and `getOperationalDependencies()`).
+        then the aggregate status' of its dependencies (see `aggregate_status()`
+        and `get_operational_dependencies()`).
 
-        If the localStatus is non-operational, that takes precedence.
-        Otherwise the localStatus is compared with its aggregate dependent status
+        If the local_status is non-operational, that takes precedence.
+        Otherwise the local_status is compared with its aggregate dependent status
         and worser value is choosen.
         """
-        if self.manualOverideStatus is not None:
-            status = self.manualOverideStatus
+        if self.manual_overide_status is not None:
+            status = self.manual_overide_status
             if status >= Status.error:
                 return status
         else:
-            status = self.localStatus
+            status = self.local_status
 
         if not status:
             return Status.unknown
@@ -98,9 +107,9 @@ class Operational(ChangeAware):
             # return error, pending, or absent
             return status
 
-        dependentStatus = self.aggregateStatus(self.getOperationalDependencies())
+        dependentStatus = self.aggregate_status(self.get_operational_dependencies())
         if dependentStatus is None:
-            # note if localStatus is a no op (status purely determined by dependents)
+            # note if local_status is a no op (status purely determined by dependents)
             # it should be set to ok
             # then ok + None = ok, which makes sense, since it has no dependendents
             return status  # return local status
@@ -118,32 +127,32 @@ class Operational(ChangeAware):
     # The computed attribute `lastChange` is whichever of the above two attributes are more recent.
 
     @property
-    def lastStateChange(self):
+    def last_state_change(self):
         return None
 
     @property
-    def lastConfigChange(self):
+    def last_config_change(self):
         return None
 
     @property
-    def lastChange(self):
-        if not self.lastStateChange:
-            return self.lastConfigChange
-        elif not self.lastConfigChange:
-            return self.lastStateChange
+    def last_change(self):
+        if not self.last_state_change:
+            return self.last_config_change
+        elif not self.last_config_change:
+            return self.last_state_change
         else:
-            return max(self.lastStateChange, self.lastConfigChange)
+            return max(self.last_state_change, self.last_config_change)
 
-    def hasChanged(self, changeset):
+    def has_changed(self, changeset):
         # if changed since the last time we checked
-        if not self.lastChange:
+        if not self.last_change:
             return False
         if not changeset:
             return True
-        return self.lastChange > changeset.changeId
+        return self.last_change > changeset.changeId
 
     @staticmethod
-    def aggregateStatus(statuses):
+    def aggregate_status(statuses):
         """
         Returns: ok, degraded, pending or None
 
@@ -204,9 +213,9 @@ class OperationalInstance(Operational):
             self._state = status._state
             self.dependencies = status.dependencies
         else:
-            self._localStatus = toEnum(Status, status)
-            self._manualOverideStatus = toEnum(Status, manualOveride)
-            self._priority = toEnum(Priority, priority)
+            self._localStatus = to_enum(Status, status)
+            self._manualOverideStatus = to_enum(Status, manualOveride)
+            self._priority = to_enum(Priority, priority)
             self._lastStateChange = lastStateChange
             self._lastConfigChange = lastConfigChange
             self._state = state
@@ -214,11 +223,11 @@ class OperationalInstance(Operational):
         # self.repairable = False # XXX3
         # self.messages = [] # XXX3
 
-    def getOperationalDependencies(self):
+    def get_operational_dependencies(self):
         return self.dependencies
 
-    def localStatus():
-        doc = "The localStatus property."
+    def local_status():
+        doc = "The local_status property."
 
         def fget(self):
             return self._localStatus
@@ -231,9 +240,9 @@ class OperationalInstance(Operational):
 
         return locals()
 
-    localStatus = property(**localStatus())
+    local_status = property(**local_status())
 
-    def manualOverideStatus():
+    def manual_overide_status():
         doc = "The manualOverideStatus property."
 
         def fget(self):
@@ -247,7 +256,7 @@ class OperationalInstance(Operational):
 
         return locals()
 
-    manualOverideStatus = property(**manualOverideStatus())
+    manual_overide_status = property(**manual_overide_status())
 
     def priority():
         doc = "The priority property."
@@ -266,11 +275,11 @@ class OperationalInstance(Operational):
     priority = property(**priority())
 
     @property
-    def lastStateChange(self):
+    def last_state_change(self):
         return self._lastStateChange
 
     @property
-    def lastConfigChange(self):
+    def last_config_change(self):
         return self._lastConfigChange
 
     def state():
@@ -280,25 +289,25 @@ class OperationalInstance(Operational):
             return self._state
 
         def fset(self, value):
-            self._state = toEnum(NodeState, value)
+            self._state = to_enum(NodeState, value)
 
         return locals()
 
     state = property(**state())
 
 
-class _ChildResources(collections.Mapping):
+class _ChildResources(Mapping):
     def __init__(self, resource):
         self.resource = resource
 
     def __getitem__(self, key):
-        return self.resource.findResource(key)
+        return self.resource.find_resource(key)
 
     def __iter__(self):
-        return iter(r.name for r in self.resource.getSelfAndDescendents())
+        return iter(r.name for r in self.resource.get_self_and_descendents())
 
     def __len__(self):
-        return len(tuple(self.resource.getSelfAndDescendents()))
+        return len(tuple(self.resource.get_self_and_descendents()))
 
 
 class EntityInstance(OperationalInstance, ResourceRef):
@@ -334,15 +343,15 @@ class EntityInstance(OperationalInstance, ResourceRef):
 
         return Ref(expr).resolve(RefContext(self, vars=vars), wantList)
 
-    def localStatus():
-        doc = "The localStatus property."
+    def local_status():
+        doc = "The working_dir property."
 
         def fget(self):
             return self._localStatus
 
         def fset(self, value):
             if self.root.attributeManager:
-                self.root.attributeManager.setStatus(self, value)
+                self.root.attributeManager.set_status(self, value)
             self._localStatus = value
 
         def fdel(self):
@@ -350,17 +359,20 @@ class EntityInstance(OperationalInstance, ResourceRef):
 
         return locals()
 
-    localStatus = property(**localStatus())
+    local_status = property(**local_status())
 
-    def getOperationalDependencies(self):
+    def get_operational_dependencies(self):
         if self.parent and self.parent is not self.root:
             yield self.parent
 
+        for d in self.dependencies:
+            yield d
+
     @property
     def key(self):
-        return "%s::.%s::%s" % (self.parent.key, self.parentRelation[1:], self.name)
+        return f"{self.parent.key}::.{self.parentRelation[1:]}::{self.name}"
 
-    def asRef(self, options=None):
+    def as_ref(self, options=None):
         return {"ref": self.key}
 
     @property
@@ -376,15 +388,15 @@ class EntityInstance(OperationalInstance, ResourceRef):
         return self.template.type
 
     @property
-    def baseDir(self):
+    def base_dir(self):
         if self.shadow:
-            return self.shadow.baseDir
+            return self.shadow.base_dir
         else:
             return self.root._baseDir
 
     @property
     def artifacts(self):
-        return self.template.artifacts
+        return {}
 
     @property
     def attributes(self):
@@ -394,16 +406,19 @@ class EntityInstance(OperationalInstance, ResourceRef):
                 # inefficient but create a local one for now
                 self.attributeManager = AttributeManager()
             # XXX3 changes to self.attributes aren't saved
-            return self.attributeManager.getAttributes(self)
+            return self.attributeManager.get_attributes(self)
 
         # returned registered attribute or create a new one
         # attribute class getter resolves references
-        return self.root.attributeManager.getAttributes(self)
+        return self.root.attributeManager.get_attributes(self)
 
     @property
     def names(self):
         return self.attributes
 
+    def get_default_relationships(self, relation=None):
+        return self.root.get_default_relationships(relation)
+
     def __eq__(self, other):
         if self is other:
             return True
@@ -416,10 +431,10 @@ class EntityInstance(OperationalInstance, ResourceRef):
                 return True
             else:
                 return False
-        if not self.lastChange:
+        if not self.last_change:
             # only support equality if resource has a changeid
             return False
-        return self.lastChange == other.lastChange and self.key == other.key
+        return self.last_change == other.last_change and self.key == other.key
 
     def __getstate__(self):
         state = self.__dict__.copy()
@@ -433,7 +448,7 @@ class EntityInstance(OperationalInstance, ResourceRef):
         return state
 
     def __repr__(self):
-        return "%s('%s')" % (self.__class__, self.name)
+        return f"{self.__class__}('{self.name}')"
 
 
 # both have occurrences
@@ -451,7 +466,7 @@ class CapabilityInstance(EntityInstance):
             self._relationships = []
         if len(self._relationships) != len(self.template.relationships):
             # instantiate missing relationships (they are only added, never deleted)
-            instantiated = set(id(c.template) for c in self._relationships)
+            instantiated = {id(c.template) for c in self._relationships}
             for template in self.template.relationships:
                 if id(template) not in instantiated:
                     rel = RelationshipInstance(
@@ -459,7 +474,7 @@ class CapabilityInstance(EntityInstance):
                     )
                     assert rel in self._relationships
                     # find the node instance that uses this requirement
-                    sourceNode = self.root.findResource(template.source.name)
+                    sourceNode = self.root.find_resource(template.source.name)
                     if sourceNode:
                         rel.source = sourceNode
 
@@ -468,7 +483,7 @@ class CapabilityInstance(EntityInstance):
     @property
     def key(self):
         # XXX implement something like _ChildResources to enable ::name instead of [.name]
-        return "%s::.%s::[.name=%s]" % (self.parent.key, "capabilities", self.name)
+        return f"{self.parent.key}::.{'capabilities'}::[.name={self.name}]"
 
 
 class RelationshipInstance(EntityInstance):
@@ -487,24 +502,66 @@ class RelationshipInstance(EntityInstance):
     def key(self):
         # XXX implement something like _ChildResources to enable ::name instead of [.name=name]
         if self.source:
-            return "%s::.requirements::[.name=%s]" % (self.source.key, self.name)
+            return f"{self.source.key}::.requirements::[.name={self.name}]"
         elif self.parent is self.root:  # it's a default relationship
-            return "::.requirements::[.name=%s]" % self.name
+            return f"::.requirements::[.name={self.name}]"
         else:  # capability is parent
-            return "%s::.relationships::[.name=%s]" % (self.parent.key, self.name)
+            return f"{self.parent.key}::.relationships::[.name={self.name}]"
 
-    def mergeProps(self, matchfn):
+    def merge_props(self, matchfn, delete_if_none=False):
         env = {}
         capability = self.parent
-        for name, val in capability.template.findProps(capability.attributes, matchfn):
-            if val is not None:
+        for name, val in capability.template.find_props(capability.attributes, matchfn):
+            if val is None:
+                if delete_if_none:
+                    env.pop(name, None)
+            else:
                 env[name] = val
-        for name, val in self.template.findProps(self.attributes, matchfn):
-            if val is not None:
+        for name, val in self.template.find_props(self.attributes, matchfn):
+            if val is None:
+                if delete_if_none:
+                    env.pop(name, None)
+            else:
                 env[name] = val
         return env
 
 
+class ArtifactInstance(EntityInstance):
+    parentRelation = "_artifacts"
+    templateType = ArtifactSpec
+
+    def __init__(
+        self, name="", attributes=None, parent=None, template=None, status=None
+    ):
+        EntityInstance.__init__(self, name, attributes, parent, template, status)
+
+    @property
+    def base_dir(self):
+        return self.template.base_dir
+
+    @property
+    def file(self):
+        return self.template.file
+
+    @property
+    def repository(self):
+        return self.template.repository
+
+    def get_path(self, resolver=None):
+        return self.template.get_path_and_fragment(resolver)
+
+    def get_path_and_fragment(self, resolver=None, tpl=None):
+        return self.template.get_path_and_fragment(resolver, tpl)
+
+    def as_import_spec(self):
+        return self.template.as_import_spec()
+
+    def get_operational_dependencies(self):
+        # skip dependency on the parent
+        for d in self.dependencies:
+            yield d
+
+
 class NodeInstance(EntityInstance):
     templateType = NodeSpec
     parentRelation = "instances"
@@ -514,15 +571,16 @@ class NodeInstance(EntityInstance):
     ):
         if parent:
             # only node instances have unique names
-            if parent.root.findResource(name):
+            if parent.root.find_resource(name):
                 raise UnfurlError(
-                    'can not create node instance "%s", its name is already in use'
-                    % name
+                    f'can not create node instance "{name}", its name is already in use'
                 )
 
-        # next two maybe initialized either by Manifest.createNodeInstance or by its property
+        # next three may be initialized either by Manifest.createNodeInstance or by its property
         self._capabilities = []
         self._requirements = []
+        self._artifacts = []
+        self._named_artifacts = None
         self.instances = []
         EntityInstance.__init__(self, name, attributes, parent, template, status)
 
@@ -532,20 +590,19 @@ class NodeInstance(EntityInstance):
 
         self._interfaces = {}
         # preload
-        self.getInterface("inherit")
-        self.getInterface("default")
+        self.get_interface("inherit")
+        self.get_interface("default")
 
-    def _findRelationship(self, relationship):
+    def _find_relationship(self, relationship):
         """
         Find RelationshipInstance that has the give relationship template
         """
         assert relationship and relationship.capability and relationship.target
         # find the Capability instance that corresponds to this relationship template's capability
-        targetNodeInstance = self.root.findResource(relationship.target.name)
-        assert targetNodeInstance, (
-            "target instance %s should have been already created"
-            % relationship.target.name
-        )
+        targetNodeInstance = self.root.find_resource(relationship.target.name)
+        assert (
+            targetNodeInstance
+        ), f"target instance {relationship.target.name} should have been already created"
         for cap in targetNodeInstance.capabilities:
             if cap.template is relationship.capability:
                 for relInstance in cap.relationships:
@@ -558,15 +615,14 @@ class NodeInstance(EntityInstance):
         # if self._requirements is None:
         if len(self._requirements) != len(self.template.requirements):
             # instantiate missing relationships (they are only added, never deleted)
-            instantiated = set(id(r.template) for r in self._requirements)
+            instantiated = {id(r.template) for r in self._requirements}
             for name, template in self.template.requirements.items():
                 assert template.relationship
                 if id(template.relationship) not in instantiated:
-                    relInstance = self._findRelationship(template.relationship)
+                    relInstance = self._find_relationship(template.relationship)
                     if not relInstance:
                         raise UnfurlError(
-                            'can not find relation instance for requirement "%s" on node "%s"'
-                            % (name, self.name)
+                            f'can not find relation instance for requirement "{name}" on node "{self.name}"'
                         )
                     assert template.relationship is relInstance.template
                     assert self.template is relInstance.template.source
@@ -574,21 +630,25 @@ class NodeInstance(EntityInstance):
 
         return self._requirements
 
-    def getRequirements(self, match):
+    def get_requirements(self, match):
+        if match is None:
+            return self.requirements
         if isinstance(match, six.string_types):
             return [r for r in self.requirements if r.template.name == match]
         elif isinstance(match, NodeInstance):
             return [r for r in self.requirements if r.target == match]
         elif isinstance(match, CapabilityInstance):
             return [r for r in self.requirements if r.parent == match]
+        elif isinstance(match, ArtifactInstance):
+            return []
         else:
-            raise UnfurlError('invalid match for getRequirements: "%s"' % match)
+            raise UnfurlError(f'invalid match for get_requirements: "{match}"')
 
     @property
     def capabilities(self):
         if len(self._capabilities) != len(self.template.capabilities):
             # instantiate missing capabilities (they are only added, never deleted)
-            instantiated = set(id(c.template) for c in self._capabilities)
+            instantiated = {id(c.template) for c in self._capabilities}
             for name, template in self.template.capabilities.items():
                 if id(template) not in instantiated:
                     # if capabilities:
@@ -602,23 +662,40 @@ class NodeInstance(EntityInstance):
 
         return self._capabilities
 
-    def getCapabilities(self, name):
+    def get_capabilities(self, name):
         return [
             capability
             for capability in self.capabilities
             if capability.template.name == name
         ]
 
-    def _getDefaultRelationships(self, relation=None):
+    @property
+    def artifacts(self) -> dict:
+        # only include named artifacts
+        if self._named_artifacts is None:
+            self._named_artifacts = {}
+            instantiated = {a.name: a for a in self._artifacts}
+            for name, template in self.template.artifacts.items():
+                artifact = instantiated.get(name)
+                if not artifact:
+                    artifact = ArtifactInstance(
+                        template.name, parent=self, template=template
+                    )
+                    assert artifact in self._artifacts
+
+                self._named_artifacts[template.name] = artifact
+        return self._named_artifacts
+
+    def _get_default_relationships(self, relation=None):
         if self.root is self:
             return
-        for rel in self.root.getDefaultRelationships(relation):
+        for rel in self.root.get_default_relationships(relation):
             for capability in self.capabilities:
                 if rel.template.matches_target(capability.template):
                     yield rel
 
-    def getDefaultRelationships(self, relation=None):
-        return list(self._getDefaultRelationships(relation))
+    def get_default_relationships(self, relation=None):
+        return list(self._get_default_relationships(relation))
 
     @property
     def names(self):
@@ -650,19 +727,17 @@ class NodeInstance(EntityInstance):
 
     @property
     def key(self):
-        return "::%s" % self.name
+        return f"::{self.name}"
 
-    def getOperationalDependencies(self):
-        parent = None
-        if self.parent and self.parent is not self.root:
-            parent = self.parent
-            yield parent
+    def get_operational_dependencies(self):
+        for dep in super().get_operational_dependencies():
+            yield dep
 
         for instance in self.requirements:
-            if instance is not parent:
+            if instance is not self.parent:
                 yield instance
 
-    def getOperationalDependents(self):
+    def get_operational_dependents(self):
         seen = set()
         for cap in self.capabilities:
             for rel in cap.relationships:
@@ -676,42 +751,45 @@ class NodeInstance(EntityInstance):
                 seen.add(id(instance))
                 yield instance
 
-    def getSelfAndDescendents(self):
+    def get_self_and_descendents(self):
         "Recursive descendent including self"
         yield self
         for r in self.instances:
-            for descendent in r.getSelfAndDescendents():
+            for descendent in r.get_self_and_descendents():
                 yield descendent
 
     @property
     def descendents(self):
-        return list(self.getSelfAndDescendents())
+        return list(self.get_self_and_descendents())
 
-    def findResource(self, resourceid):
+    # XXX use find_instance instead and remove find_resource
+    def find_resource(self, resourceid):
         if self.name == resourceid:
             return self
         for r in self.instances:
-            child = r.findResource(resourceid)
+            child = r.find_resource(resourceid)
             if child:
                 return child
         return None
 
-    def findInstanceOrExternal(self, resourceid):
-        instance = self.findResource(resourceid)
+    find_instance = find_resource
+
+    def find_instance_or_external(self, resourceid):
+        instance = self.find_resource(resourceid)
         if instance:
             return instance
         if self.imports:
-            return self.imports.findImport(resourceid)
+            return self.imports.find_import(resourceid)
         return None
 
-    def addInterface(self, klass, name=None):
+    def add_interface(self, klass, name=None):
         if not isinstance(klass, six.string_types):
             klass = klass.__module__ + "." + klass.__name__
         current = self._attributes.setdefault(".interfaces", {})
         current[name or klass] = klass
         return current
 
-    def getInterface(self, name):
+    def get_interface(self, name):
         # XXX uses TOSCA interfaces instead
         if ".interfaces" not in self._attributes:
             return None  # no interfaces
@@ -724,27 +802,28 @@ class NodeInstance(EntityInstance):
         else:
             className = self._attributes[".interfaces"].get(name)
             if className:
-                instance = loadClass(className)(name, self)
+                instance = load_class(className)(name, self)
                 self._interfaces[name] = instance
                 return instance
         return None
 
     def __repr__(self):
-        return "NodeInstance('%s')" % self.name
+        return f"NodeInstance('{self.name}')"
 
 
 class TopologyInstance(NodeInstance):
     templateType = TopologySpec
 
     def __init__(self, template, status=None):
-        NodeInstance.__init__(self, "root", template=template, status=status)
-        # add these as special child resources so they can be accessed like "::inputs::foo"
-        self.inputs = NodeInstance("inputs", template.inputs, self)
-        self.outputs = NodeInstance("outputs", template.outputs, self)
+        attributes = dict(inputs=template.inputs, outputs=template.outputs)
+        NodeInstance.__init__(
+            self, "root", attributes, template=template, status=status
+        )
+
         self._relationships = None
         self._tmpDir = None
 
-    def setBaseDir(self, baseDir):
+    def set_base_dir(self, baseDir):
         self._baseDir = baseDir
         if not self._templar or self._templar._basedir != baseDir:
             loader = DataLoader()
@@ -767,21 +846,22 @@ class TopologyInstance(NodeInstance):
 
         return self._relationships
 
-    def getDefaultRelationships(self, relation=None):
+    def get_default_relationships(self, relation=None):
         # for root, this is the same as self.requirements
         if not relation:
             return self.requirements
         return [
-            rel for rel in self.requirements if rel.template.isCompatibleType(relation)
+            rel
+            for rel in self.requirements
+            if rel.template.is_compatible_type(relation)
         ]
 
-    def getOperationalDependencies(self):
+    def get_operational_dependencies(self):
         for instance in self.instances:
-            if instance.name not in ["inputs", "outputs"]:
-                yield instance
+            yield instance
 
     @property
-    def tmpDir(self):
+    def tmp_dir(self):
         if not self._tmpDir:
-            self._tmpDir = makeTempDir()
+            self._tmpDir = make_temp_dir()
         return self._tmpDir
