@@ -11,14 +11,7 @@ import re
 import os
 import fnmatch
 import shutil
-import collections
-
-if os.name == "posix" and sys.version_info[0] < 3:
-    import subprocess32 as subprocess
-else:
-    import subprocess
-
-from collections import Mapping
+from collections.abc import Mapping, MutableSequence
 import os.path
 from jsonschema import Draft7Validator, validators, RefResolver
 import jsonschema.exceptions
@@ -30,6 +23,8 @@ from ansible.utils.unsafe_proxy import AnsibleUnsafeText, AnsibleUnsafeBytes, wr
 import warnings
 import codecs
 import io
+from contextlib import contextmanager
+from logging import Logger
 
 try:
     from shutil import which
@@ -50,7 +45,7 @@ try:
     imp = None
 except ImportError:
     import imp
-from . import sensitive
+from .logs import sensitive
 import logging
 
 logger = logging.getLogger("unfurl")
@@ -65,12 +60,12 @@ except ImportError:
 _basepath = os.path.abspath(os.path.dirname(__file__))
 
 
-def getPackageDigest():
+def get_package_digest():
     from git import Repo
 
     basedir = os.path.dirname(_basepath)
     if os.path.isdir(os.path.join(basedir, ".git")):
-        return Repo(basedir).git.describe("--dirty")
+        return Repo(basedir).git.describe("--dirty", "--always")
 
     try:
         pbr = [p for p in files("unfurl") if "pbr.json" in str(p)][0]
@@ -85,26 +80,26 @@ class UnfurlError(Exception):
             (etype, value, traceback) = sys.exc_info()
             if value:
                 message = str(message) + ": " + str(value)
-        super(UnfurlError, self).__init__(message)
+        super().__init__(message)
         self.stackInfo = (etype, value, traceback) if saveStack and value else None
         if log:
             logger.error(message, exc_info=True)
 
-    def getStackTrace(self):
+    def get_stack_trace(self):
         if not self.stackInfo:
             return ""
         return "".join(traceback.format_exception(*self.stackInfo))
 
 
 class UnfurlValidationError(UnfurlError):
-    def __init__(self, message, errors=None):
-        super(UnfurlValidationError, self).__init__(message)
+    def __init__(self, message, errors=None, log=False):
+        super().__init__(message, log=log)
         self.errors = errors or []
 
 
 class UnfurlTaskError(UnfurlError):
     def __init__(self, task, message, log=logging.ERROR):
-        super(UnfurlTaskError, self).__init__(message, True, False)
+        super().__init__(message, True, False)
         self.task = task
         task._errors.append(self)
         self.severity = log
@@ -114,12 +109,12 @@ class UnfurlTaskError(UnfurlError):
 
 class UnfurlAddingResourceError(UnfurlTaskError):
     def __init__(self, task, resourceSpec, name, log=logging.DEBUG):
-        message = "error updating resource %s" % name
-        super(UnfurlAddingResourceError, self).__init__(task, message, log)
+        message = f"error updating resource {name}"
+        super().__init__(task, message, log)
         self.resourceSpec = resourceSpec
 
 
-def wrapSensitiveValue(obj, vault=None):
+def wrap_sensitive_value(obj):
     # we don't remember the vault and vault id associated with this value
     # so the value will be rekeyed with whichever vault is associated with the serializing yaml
     if isinstance(obj, bytes):
@@ -132,24 +127,24 @@ def wrapSensitiveValue(obj, vault=None):
                 return sensitive_bytes(obj)
     elif isinstance(obj, six.string_types):
         return sensitive_str(obj)
-    elif isinstance(obj, collections.Mapping):
+    elif isinstance(obj, Mapping):
         return sensitive_dict(obj)
-    elif isinstance(obj, collections.MutableSequence):
+    elif isinstance(obj, MutableSequence):
         return sensitive_list(obj)
     else:
         return obj
 
 
-def isSensitive(obj):
+def is_sensitive(obj):
     test = getattr(obj, "__sensitive__", None)
     if test:
         return test()
     if isinstance(obj, AnsibleVaultEncryptedUnicode):
         return True
-    elif isinstance(obj, collections.Mapping):
-        return any(isSensitive(i) for i in obj.values())
-    elif isinstance(obj, collections.MutableSequence):
-        return any(isSensitive(i) for i in obj)
+    elif isinstance(obj, Mapping):
+        return any(is_sensitive(i) for i in obj.values())
+    elif isinstance(obj, MutableSequence):
+        return any(is_sensitive(i) for i in obj)
     return False
 
 
@@ -159,7 +154,7 @@ class sensitive_bytes(AnsibleUnsafeBytes, sensitive):
 
     def decode(self, *args, **kwargs):
         """Wrapper method to ensure type conversions maintain sensitive context"""
-        return sensitive_str(super(sensitive_bytes, self).decode(*args, **kwargs))
+        return sensitive_str(super().decode(*args, **kwargs))
 
 
 # also mark AnsibleUnsafe so wrap_var doesn't strip out sensitive status
@@ -168,7 +163,7 @@ class sensitive_str(AnsibleUnsafeText, sensitive):
 
     def encode(self, *args, **kwargs):
         """Wrapper method to ensure type conversions maintain sensitive context"""
-        return sensitive_bytes(super(sensitive_str, self).encode(*args, **kwargs))
+        return sensitive_bytes(super().encode(*args, **kwargs))
 
 
 class sensitive_dict(dict, sensitive):
@@ -179,19 +174,21 @@ class sensitive_list(list, sensitive):
     """Transparent wrapper class to mark a list as sensitive"""
 
 
-def toYamlText(val):
+def to_yaml_text(val):
     if isinstance(val, (ScalarString, sensitive)):
         return val
     # convert or copy string (copy to deal with things like AnsibleUnsafeText)
     val = str(val)
     if "\n" in val:
+        if six.PY2 and isinstance(val, str):
+            val = val.decode("utf-8")
         return FoldedScalarString(val)
     return val
 
 
-def assertForm(src, types=Mapping, test=True):
+def assert_form(src, types=Mapping, test=True):
     if not isinstance(src, types) or not test:
-        raise UnfurlError("Malformed definition: %s" % src)
+        raise UnfurlError(f"Malformed definition: {src}")
     return src
 
 
@@ -199,20 +196,20 @@ _ClassRegistry = {}
 _shortNameRegistry = {}
 
 
-def registerShortNames(shortNames):
+def register_short_names(shortNames):
     _shortNameRegistry.update(shortNames)
 
 
-def registerClass(className, factory, shortName=None, replace=True):
-    if shortName:
-        _shortNameRegistry[shortName] = className
+def register_class(className, factory, short_name=None, replace=True):
+    if short_name:
+        _shortNameRegistry[short_name] = className
     if not replace and className in _ClassRegistry:
         if _ClassRegistry[className] is not factory:
-            raise UnfurlError("class already registered for %s" % className)
+            raise UnfurlError(f"class already registered for {className}")
     _ClassRegistry[className] = factory
 
 
-def loadModule(path, full_name=None):
+def load_module(path, full_name=None):
     if full_name is None:
         full_name = re.sub(r"\W", "_", path)  # generate a name from the path
     if full_name in sys.modules:
@@ -220,10 +217,8 @@ def loadModule(path, full_name=None):
 
     with warnings.catch_warnings():
         warnings.simplefilter("ignore", RuntimeWarning)
-        if imp is None:
-            spec = importlib.util.spec_from_file_location(
-                to_native(full_name), to_native(path)
-            )
+        if imp is None:  # Python 3
+            spec = importlib.util.spec_from_file_location(full_name, path)
             module = importlib.util.module_from_spec(spec)
             spec.loader.exec_module(module)
             sys.modules[full_name] = module
@@ -236,7 +231,7 @@ def loadModule(path, full_name=None):
     return module
 
 
-def loadClass(klass, defaultModule="__main__"):
+def load_class(klass, defaultModule="__main__"):
     import importlib
 
     prefix, sep, suffix = klass.rpartition(".")
@@ -247,7 +242,7 @@ def loadClass(klass, defaultModule="__main__"):
 _shortNameRegistry = {}
 
 
-def lookupClass(kind, apiVersion=None, default=None):
+def lookup_class(kind, apiVersion=None, default=None):
     if kind in _ClassRegistry:
         return _ClassRegistry[kind]
     elif kind in _shortNameRegistry:
@@ -255,16 +250,16 @@ def lookupClass(kind, apiVersion=None, default=None):
     else:
         className = kind
     try:
-        klass = loadClass(className)
+        klass = load_class(className)
     except ImportError:
         klass = None
 
     if klass:
-        registerClass(className, klass)
+        register_class(className, klass)
     return klass
 
 
-def toEnum(enum, value, default=None):
+def to_enum(enum, value, default=None):
     # from string: Status[name]; to string: status.name
     if isinstance(value, six.string_types):
         return enum[value]
@@ -315,30 +310,33 @@ def dump(obj, tp, suffix="", yaml=None, encoding=None):
             f.detach()
 
 
-def _saveToVault(path, obj, yaml=None, encoding=None, fd=None):
+def _save_to_vault(path, obj, yaml=None, encoding=None, fd=None):
     vaultExt = path.endswith(".vault")
     if vaultExt or encoding == "vault":
-        assert yaml and yaml.representer.vault
-        vault = VaultEditor(yaml.representer.vault)
+        assert yaml and yaml.representer.vault and yaml.representer.vault.secrets
+        vaultlib = yaml.representer.vault
+        vault = VaultEditor(vaultlib)
         f = io.BytesIO()
         vpath = path[: -len(".vault")] if vaultExt else path
         dump(obj, f, vpath, yaml, encoding)
-        b_vaulttext = yaml.representer.vault.encrypt(f.getvalue())
+        # the first vaultid is the most specific to the current project so encrypt with that one
+        vault_id, secret = vaultlib.secrets[0]
+        b_vaulttext = vaultlib.encrypt(f.getvalue(), secret=secret, vault_id=vault_id)
         vault.write_data(b_vaulttext, fd or path)
         return True
     return False
 
 
-def saveToFile(path, obj, yaml=None, encoding=None):
+def save_to_file(path, obj, yaml=None, encoding=None):
     dir = os.path.dirname(path)
     if dir and not os.path.isdir(dir):
         os.makedirs(dir)
-    if not _saveToVault(path, obj, yaml, encoding):
+    if not _save_to_vault(path, obj, yaml, encoding):
         with open(path, "wb") as f:
             dump(obj, f, path, yaml, encoding)
 
 
-def saveToTempfile(obj, suffix="", delete=True, dir=None, yaml=None, encoding=None):
+def save_to_tempfile(obj, suffix="", delete=True, dir=None, yaml=None, encoding=None):
     tp = tempfile.NamedTemporaryFile(
         "w+b",
         suffix=suffix or "",
@@ -348,7 +346,7 @@ def saveToTempfile(obj, suffix="", delete=True, dir=None, yaml=None, encoding=No
     if delete:
         atexit.register(lambda: os.path.exists(tp.name) and os.unlink(tp.name))
 
-    if not _saveToVault(suffix or "", obj, yaml, encoding, tp.fileno()):
+    if not _save_to_vault(suffix or "", obj, yaml, encoding, tp.fileno()):
         try:
             dump(obj, tp, suffix or "", yaml, encoding)
         finally:
@@ -356,14 +354,14 @@ def saveToTempfile(obj, suffix="", delete=True, dir=None, yaml=None, encoding=No
     return tp
 
 
-def makeTempDir(delete=True, prefix="unfurl"):
+def make_temp_dir(delete=True, prefix="unfurl"):
     tempDir = tempfile.mkdtemp(prefix, dir=os.environ.get("UNFURL_TMPDIR"))
     if delete:
         atexit.register(lambda: os.path.isdir(tempDir) and shutil.rmtree(tempDir))
     return tempDir
 
 
-def getBaseDir(path):
+def get_base_dir(path):
     if os.path.exists(path):
         isdir = os.path.isdir(path)
     else:
@@ -374,6 +372,22 @@ def getBaseDir(path):
         return os.path.normpath(os.path.dirname(path))
 
 
+@contextmanager
+def change_cwd(new_path: str, log: Logger = None):
+    """Temporarily change current working directory"""
+    old_path = os.getcwd()
+    if new_path:
+        if log:
+            log.trace("Changing CWD to: %s", new_path)
+        os.chdir(new_path)
+    try:
+        yield
+    finally:
+        if log:
+            log.trace("Restoring CWD back to: %s", old_path)
+        os.chdir(old_path)
+
+
 # https://python-jsonschema.readthedocs.io/en/latest/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance
 # XXX unused because this breaks check_schema
 def extend_with_default(validator_class):
@@ -408,11 +422,11 @@ DefaultValidatingLatestDraftValidator = (
 )
 
 
-def validateSchema(obj, schema, baseUri=None):
-    return not findSchemaErrors(obj, schema)
+def validate_schema(obj, schema, baseUri=None):
+    return not find_schema_errors(obj, schema)
 
 
-def findSchemaErrors(obj, schema, baseUri=None):
+def find_schema_errors(obj, schema, baseUri=None):
     # XXX2 have option that includes definitions from manifest's schema
     if baseUri is not None:
         resolver = RefResolver(base_uri=baseUri, referrer=schema)
@@ -424,7 +438,10 @@ def findSchemaErrors(obj, schema, baseUri=None):
     error = jsonschema.exceptions.best_match(errors)
     if not error:
         return None
-    message = "%s in %s" % (error.message, "/".join(error.absolute_path))
+    message = "%s in %s" % (
+        error.message,
+        "/".join([str(p) for p in error.absolute_path]),
+    )
     return message, errors
 
 
@@ -468,7 +485,7 @@ class ChainMap(Mapping):
         return "ChainMap(%r)" % (self._maps,)
 
 
-class Generate(object):
+class Generate:
     """
     Roughly equivalent to "yield from" but works in Python < 3.3
 
@@ -492,24 +509,33 @@ class Generate(object):
             return False
 
 
-def filterEnv(rules, env=None, addOnly=False):
+def filter_env(rules, env=None, addOnly=False):
     """
-    If 'env' is None it will be set to os.environ
+    Applies the given list of rules to a dictionary of environment variables and returns a new dictionary.
 
-    If addOnly is False (the default) all variables in `env` will be included
-    in the returned dict, otherwise only variables added by `rules` will be included
+    Args:
+        rules (dict): A dictionary of rules for adding, removing and filtering environment variables.
+        env (dict, optional): The environment to apply the give rules to. If ``env`` is None it will be set to ``os.environ``.
+          Defaults to None.
+        addOnly (bool, optional): If addOnly is False (the default) all variables in ``env`` will be included
+          in the returned dict, otherwise only variables added by ``rules`` will be included
 
-    foo: bar # add foo=bar
-    +foo: # copy foo
-    +foo: bar # copy foo, set it bar if not present
-    +!foo*: # copy all except keys matching "foo*"
-    -!foo: # remove all except foo
-    ^foo: /bar/bin # treat foo like a PATH and prepend value: /bar/bin:$foo
+    Rules applied in the order they are declared in the ``rules`` dictionary. The following examples show the different patterns for the rules:
+
+        :foo \: bar: Add ``foo=bar``
+        :+foo:  Copy ``foo`` from the current environment
+        :+foo \: bar: Copy ``foo``, or add ``foo=bar`` if it is not present
+        :+!foo*: Copy all name from the current environment except those matching ``foo*``
+        :-!foo:  Remove all names except for ``foo``
+        :^foo \: /bar/bin: Treat ``foo`` like ``PATH`` and prepend ``/bar/bin:$foo``
     """
     if env is None:
         env = os.environ
 
-    start = {} if addOnly else env.copy()
+    if addOnly:
+        start = {}
+    else:
+        start = env.copy()
     for name, val in rules.items():
         if name[:1] in "+-^":
             # add or remove from env
@@ -540,7 +566,7 @@ def filterEnv(rules, env=None, addOnly=False):
                     start.update(match)
                 elif val is not None:  # name isn't in existing, use default is set
                     start[name] = val
-        else:
+        elif val is not None:  # don't set if val is None
             start[name] = val
     return start
 
@@ -559,11 +585,13 @@ required_envvars = [
     "HTTPS_PROXY",
     "NO_PROXY",
     "PYTHONPATH",
+    "VIRTUAL_ENV",
     "UNFURL_TMPDIR",
     "UNFURL_LOGGING",
     "UNFURL_HOME",
     "UNFURL_RUNTIME",
     "UNFURL_NORUNTIME",
+    "UNFURL_APPROVE",
 ]
 # hack for sphinx ext documentedlist
 _sphinx_envvars = [(i,) for i in required_envvars]
