@@ -5,6 +5,7 @@ import sys
 import codecs
 import json
 import functools
+from typing import Union
 import six
 from six.moves import urllib
 
@@ -27,20 +28,20 @@ from .util import (
     sensitive_bytes,
     sensitive_dict,
     sensitive_list,
-    wrapSensitiveValue,
+    wrap_sensitive_value,
     UnfurlError,
     UnfurlValidationError,
-    findSchemaErrors,
-    getBaseDir,
+    find_schema_errors,
+    get_base_dir,
 )
 from .merge import (
-    expandDoc,
-    makeMapWithBase,
-    findAnchor,
-    _cacheAnchors,
-    restoreIncludes,
+    expand_doc,
+    make_map_with_base,
+    find_anchor,
+    _cache_anchors,
+    restore_includes,
 )
-from .repo import isURLorGitPath
+from .repo import is_url_or_git_path
 from toscaparser.common.exception import URLException, ExceptionCollector
 from toscaparser.utils.gettextutils import _
 import toscaparser.imports
@@ -53,11 +54,16 @@ import logging
 
 logger = logging.getLogger("unfurl")
 
+CLEARTEXT_VAULT = VaultLib(secrets=[("cleartext", b"")])
+
+
+def _use_clear_text(vault):
+    clear_id = CLEARTEXT_VAULT.secrets[0][0]
+    return vault.secrets and all(s[0] == clear_id for s in vault.secrets)
+
 
 def represent_undefined(self, data):
-    raise RepresenterError(
-        "cannot represent an object: <%s> of type %s" % (data, type(data))
-    )
+    raise RepresenterError(f"cannot represent an object: <{data}> of type {type(data)}")
 
 
 def _represent_sensitive(dumper, data, tag):
@@ -66,22 +72,32 @@ def _represent_sensitive(dumper, data, tag):
         return dumper.represent_scalar(tag, b_ciphertext.decode(), style="|")
     else:
         return dumper.represent_scalar(
-            u"tag:yaml.org,2002:str", sensitive_str.redacted_str
+            "tag:yaml.org,2002:str", sensitive_str.redacted_str
         )
 
 
 def represent_sensitive_str(dumper, data):
-    return _represent_sensitive(dumper, data, u"!vault")
+    if _use_clear_text(dumper.vault):
+        return SafeRepresenter.represent_str(dumper, data)
+    return _represent_sensitive(dumper, data, "!vault")
 
 
-def represent_sensitive_json(dumper, data):
-    return _represent_sensitive(
-        dumper, json.dumps(data, sort_keys=True), u"!vault-json"
-    )
+def represent_sensitive_list(dumper, data):
+    if _use_clear_text(dumper.vault):
+        return SafeRepresenter.represent_list(dumper, data)
+    return _represent_sensitive(dumper, json.dumps(data, sort_keys=True), "!vault-json")
+
+
+def represent_sensitive_dict(dumper, data):
+    if _use_clear_text(dumper.vault):
+        return SafeRepresenter.represent_dict(dumper, data)
+    return _represent_sensitive(dumper, json.dumps(data, sort_keys=True), "!vault-json")
 
 
 def represent_sensitive_bytes(dumper, data):
-    return _represent_sensitive(dumper, data, u"!vault-binary")
+    if _use_clear_text(dumper.vault):
+        return SafeRepresenter.represent_binary(dumper, data)
+    return _represent_sensitive(dumper, data, "!vault-binary")
 
 
 def _construct_vault(constructor, node, tag):
@@ -90,7 +106,7 @@ def _construct_vault(constructor, node, tag):
         raise ConstructorError(
             context=None,
             context_mark=None,
-            problem="found %s but no vault password provided" % tag,
+            problem=f"found {tag} but no vault password provided",
             problem_mark=node.start_mark,
             note=None,
         )
@@ -106,7 +122,7 @@ def construct_vault(constructor, node):
 def construct_vaultjson(constructor, node):
     value = _construct_vault(constructor, node, "!vault-json")
     cleartext = to_text(constructor.vault.decrypt(value))
-    return wrapSensitiveValue(json.loads(cleartext))
+    return wrap_sensitive_value(json.loads(cleartext))
 
 
 def construct_vaultbinary(constructor, node):
@@ -115,7 +131,7 @@ def construct_vaultbinary(constructor, node):
     return sensitive_bytes(cleartext)
 
 
-def makeYAML(vault=None):
+def make_yaml(vault=None):
     if not vault:
         vault = VaultLib(secrets=None)
     yaml = YAML()
@@ -125,32 +141,29 @@ def makeYAML(vault=None):
     yaml.representer.add_representer(None, represent_undefined)
 
     yaml.constructor.vault = vault
-    yaml.constructor.add_constructor(u"!vault", construct_vault)
-    yaml.constructor.add_constructor(u"!vault-json", construct_vaultjson)
-    yaml.constructor.add_constructor(u"!vault-binary", construct_vaultbinary)
+    yaml.constructor.add_constructor("!vault", construct_vault)
+    yaml.constructor.add_constructor("!vault-json", construct_vaultjson)
+    yaml.constructor.add_constructor("!vault-binary", construct_vaultbinary)
 
     yaml.representer.vault = vault
+    # write out <<REDACTED>> or encrypt depending on vault.secrets
     yaml.representer.add_representer(sensitive_str, represent_sensitive_str)
-    yaml.representer.add_representer(sensitive_dict, represent_sensitive_json)
-    yaml.representer.add_representer(sensitive_list, represent_sensitive_json)
+    yaml.representer.add_representer(sensitive_dict, represent_sensitive_dict)
+    yaml.representer.add_representer(sensitive_list, represent_sensitive_list)
     yaml.representer.add_representer(sensitive_bytes, represent_sensitive_bytes)
 
-    if six.PY3:
-        represent_unicode = SafeRepresenter.represent_str
-        represent_binary = SafeRepresenter.represent_binary
-    else:
-        represent_unicode = SafeRepresenter.represent_unicode
-        represent_binary = SafeRepresenter.represent_str
-
-    yaml.representer.add_representer(AnsibleUnsafeText, represent_unicode)
-    yaml.representer.add_representer(AnsibleUnsafeBytes, represent_binary)
+    yaml.representer.add_representer(AnsibleUnsafeText, SafeRepresenter.represent_str)
+    yaml.representer.add_representer(
+        AnsibleUnsafeBytes, SafeRepresenter.represent_binary
+    )
     return yaml
 
 
-yaml = makeYAML()
+yaml = make_yaml()
+cleartext_yaml = make_yaml(CLEARTEXT_VAULT)
 
 
-def makeVaultLib(passwordBytes, vaultId="default"):
+def make_vault_lib(passwordBytes, vaultId="default"):
     if passwordBytes:
         if isinstance(passwordBytes, six.string_types):
             passwordBytes = to_bytes(passwordBytes)
@@ -159,49 +172,84 @@ def makeVaultLib(passwordBytes, vaultId="default"):
     return None
 
 
+def make_vault_lib_ex(secrets: [str, Union[str, bytes]]):
+    vault_secrets = []
+    for vaultId, passwordBytes in secrets:
+        if isinstance(passwordBytes, six.string_types):
+            passwordBytes = to_bytes(passwordBytes)
+        vault_secrets.append((vaultId, VaultSecret(passwordBytes)))
+    return VaultLib(secrets=vault_secrets)
+
+
 _refResolver = RefResolver("", None)
 
 
 class ImportResolver(toscaparser.imports.ImportResolver):
-    def __init__(self, manifest, ignoreFileNotFound=False):
+    def __init__(self, manifest, ignoreFileNotFound=False, expand=False):
         self.manifest = manifest
         self.ignoreFileNotFound = ignoreFileNotFound
         self.loader = manifest.loader
+        self.expand = expand
+
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        state["loader"] = None
+        return state
+
+    def get_repository(self, name, tpl):
+        if isinstance(tpl, dict) and "url" in tpl:
+            url = tpl["url"]
+            if (
+                "://" not in url and "@" in url
+            ):  # scp style used by git: user@server:project.git
+                # convert to ssh://user@server/project.git
+                url = "ssh://" + url.replace(":", "/", 1)
+            tpl["url"] = url
+        return Repository(name, tpl)
 
     def get_url(self, importLoader, repository_name, file_name, isFile=None):
         # returns url or path, isFile, fragment
         importLoader.stream = None
         if repository_name:
-            # don't include file_name, we need to keep it distinct from the repo path till the end
-            url_info = super(ImportResolver, self).get_url(
-                importLoader, repository_name, "", isFile
-            )
+            path = self.get_repository_url(importLoader, repository_name)
+            file_name, sep, fragment = file_name.partition("#")
+            if path.startswith("file:"):
+                path = path[5:]
+                if path.startswith("//"):
+                    path = path[2:]
+                isFile = True
         else:
-            url_info = super(ImportResolver, self).get_url(
-                importLoader, repository_name, file_name, isFile
-            )
+            url_info = super().get_url(importLoader, repository_name, file_name, isFile)
+            if not url_info:
+                return url_info
             file_name = ""
-        if not url_info:
-            return url_info
+            path, isFile, fragment = url_info
 
-        path, isFile, fragment = url_info
-        if not isFile or isURLorGitPath(path):  # only support urls to git repos for now
-            repo, filePath, revision, bare = self.manifest.findRepoFromGitUrl(
+        if not isFile or is_url_or_git_path(path):
+            # only support urls to git repos for now
+            repo_view = self.manifest.repositories.get(repository_name)
+            if repo_view and repo_view.repo:
+                return os.path.join(repo_view.working_dir, file_name), True, fragment
+
+            repo, filePath, revision, bare = self.manifest.find_repo_from_git_url(
                 path, isFile, importLoader
             )
             if not repo:
                 raise UnfurlError("Could not resolve git URL: " + path)
             isFile = True
             if bare:
-                if not filePath:
+                # XXX support empty filePath or when filePath is a directory -- need to checkout the tree
+                if (
+                    not filePath
+                    or repo.repo.rev_parse(revision + ":" + filePath).type != "blob"
+                ):
                     raise UnfurlError(
-                        "can't retrieve local repository for '%s' with revision '%s'"
-                        % (path, revision)
+                        f"can't retrieve local repository for '{path}' with revision '{revision}'"
                     )
                 importLoader.stream = six.StringIO(repo.show(filePath, revision))
                 path = os.path.join(filePath, file_name or "")
             else:
-                path = os.path.join(repo.workingDir, filePath, file_name or "").rstrip(
+                path = os.path.join(repo.working_dir, filePath, file_name or "").rstrip(
                     "/"
                 )
                 repository = None
@@ -209,7 +257,7 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                     repository = Repository(
                         repository_name, importLoader.repositories[repository_name]
                     )
-                    self.manifest.addRepository(repo, repository, filePath)
+                    self.manifest.add_repository(repo, repository, filePath)
         elif file_name:
             path = os.path.join(path, file_name)
         return path, isFile, fragment
@@ -240,15 +288,19 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                 except urllib.error.URLError as e:
                     if hasattr(e, "reason"):
                         msg = _(
-                            'Failed to reach server "%(path)s". Reason is: '
-                            "%(reason)s."
+                            (
+                                'Failed to reach server "%(path)s". Reason is: '
+                                + "%(reason)s."
+                            )
                         ) % {"path": path, "reason": e.reason}
                         ExceptionCollector.appendException(URLException(what=msg))
                         return
                     elif hasattr(e, "code"):
                         msg = _(
-                            'The server "%(path)s" couldn\'t fulfill the request. '
-                            'Error code: "%(code)s".'
+                            (
+                                'The server "%(path)s" couldn\'t fulfill the request. '
+                                + 'Error code: "%(code)s".'
+                            )
                         ) % {"path": path, "code": e.code}
                         ExceptionCollector.appendException(URLException(what=msg))
                         return
@@ -257,20 +309,26 @@ class ImportResolver(toscaparser.imports.ImportResolver):
             with f:
                 doc = yaml.load(f.read())
                 if isinstance(doc, CommentedMap):
+                    if self.expand:
+                        includes, doc = expand_doc(
+                            doc, cls=make_map_with_base(doc, get_base_dir(path))
+                        )
+                        doc.includes = includes
                     doc.path = path
+
             if fragment:
                 return _refResolver.resolve_fragment(doc, fragment)
             else:
                 return doc
         except:
             if path != originalPath:
-                msg = 'Could not load "%s" (originally "%s")' % (path, originalPath)
+                msg = f'Could not load "{path}" (originally "{originalPath}")'
             else:
-                msg = 'Could not load "%s"' % path
+                msg = f'Could not load "{path}"'
             raise UnfurlError(msg, True)
 
 
-class YamlConfig(object):
+class YamlConfig:
     def __init__(
         self,
         config=None,
@@ -309,19 +367,19 @@ class YamlConfig(object):
                 self.config = config
             if not isinstance(self.config, CommentedMap):
                 raise UnfurlValidationError(
-                    'invalid YAML document with contents: "%s"' % self.config
+                    f'invalid YAML document with contents: "{self.config}"'
                 )
 
-            findAnchor(self.config, None)  # create _anchorCache
+            find_anchor(self.config, None)  # create _anchorCache
             self._cachedDocIncludes = {}
             # schema should include defaults but can't validate because it doesn't understand includes
             # but should work most of time
-            self.config.loadTemplate = self.loadInclude
+            self.config.loadTemplate = self.load_include
             self.loadHook = loadHook
 
-            self.baseDirs = [self.getBaseDir()]
-            self.includes, expandedConfig = expandDoc(
-                self.config, cls=makeMapWithBase(self.config, self.baseDirs[0])
+            self.baseDirs = [self.get_base_dir()]
+            self.includes, expandedConfig = expand_doc(
+                self.config, cls=make_map_with_base(self.config, self.baseDirs[0])
             )
             self.expanded = expandedConfig
             errors = schema and self.validate(expandedConfig)
@@ -332,35 +390,36 @@ class YamlConfig(object):
                 )
             else:
                 self.valid = not not errors
-        except:
+        except Exception as e:
             if self.path:
-                msg = "Unable to load yaml config at %s" % self.path
+                msg = f"Unable to load yaml config at {self.path}"
             else:
                 msg = "Unable to parse yaml config"
             raise UnfurlError(msg, True)
 
-    def loadYaml(self, path, baseDir=None, warnWhenNotFound=False):
-        path = os.path.abspath(os.path.join(baseDir or self.getBaseDir(), path))
+    def load_yaml(self, path, baseDir=None, warnWhenNotFound=False):
+        path = os.path.abspath(os.path.join(baseDir or self.get_base_dir(), path))
         if warnWhenNotFound and not os.path.isfile(path):
             return path, None
+        logger.debug("attempting to load YAML file: %s", path)
         with open(path, "r") as f:
             config = self.yaml.load(f)
         return path, config
 
-    def getBaseDir(self):
+    def get_base_dir(self):
         if self.path:
-            return getBaseDir(self.path)
+            return get_base_dir(self.path)
         else:
             return "."
 
-    def restoreIncludes(self, changed):
-        restoreIncludes(self.includes, self.config, changed, cls=CommentedMap)
+    def restore_includes(self, changed):
+        restore_includes(self.includes, self.config, changed, cls=CommentedMap)
 
     def dump(self, out=sys.stdout):
         try:
             self.yaml.dump(self.config, out)
         except:
-            raise UnfurlError("Error saving %s" % self.path, True)
+            raise UnfurlError(f"Error saving {self.path}", True)
 
     def save(self):
         output = six.StringIO()
@@ -369,9 +428,14 @@ class YamlConfig(object):
             if self.lastModified:
                 statinfo = os.stat(self.path)
                 if statinfo.st_mtime > self.lastModified:
+                    logger.error(
+                        'Not saving "%s", it was unexpectedly modified after it was loaded, %d is after last modified time %d',
+                        self.path,
+                        statinfo.st_mtime,
+                        self.lastModified,
+                    )
                     raise UnfurlError(
-                        'Not saving "%s", it was unexpectedly modified after it was loaded'
-                        % self.path
+                        f'Not saving "{self.path}", it was unexpectedly modified after it was loaded'
                     )
             with open(self.path, "w") as f:
                 f.write(output.getvalue())
@@ -382,7 +446,7 @@ class YamlConfig(object):
     @property
     def yaml(self):
         if not self._yaml:
-            self._yaml = makeYAML(self.vault)
+            self._yaml = make_yaml(self.vault)
         return self._yaml
 
     def __getstate__(self):
@@ -403,9 +467,9 @@ class YamlConfig(object):
         baseUri = None
         if path:
             baseUri = urljoin("file:", pathname2url(path))
-        return findSchemaErrors(config, self.schema, baseUri)
+        return find_schema_errors(config, self.schema, baseUri)
 
-    def searchIncludes(self, key=None, pathPrefix=None):
+    def search_includes(self, key=None, pathPrefix=None):
         for k in self._cachedDocIncludes:
             path, template = self._cachedDocIncludes[k]
             candidate = True
@@ -417,17 +481,17 @@ class YamlConfig(object):
                 return k, template
         return None, None
 
-    def saveInclude(self, key):
+    def save_include(self, key):
         path, template = self._cachedDocIncludes[key]
         output = six.StringIO()
         try:
             self.yaml.dump(template, output)
         except:
-            raise UnfurlError("Error saving include %s" % path, True)
+            raise UnfurlError(f"Error saving include {path}", True)
         with open(path, "w") as f:
             f.write(output.getvalue())
 
-    def loadInclude(
+    def load_include(
         self, templatePath, warnWhenNotFound=False, expanded=None, check=False
     ):
         if check:
@@ -451,7 +515,7 @@ class YamlConfig(object):
             value = templatePath.get("merge")
             if "file" not in templatePath:
                 raise UnfurlError(
-                    "`file` key missing from document include: %s" % templatePath
+                    f"`file` key missing from document include: {templatePath}"
                 )
             key = templatePath["file"]
         else:
@@ -471,23 +535,20 @@ class YamlConfig(object):
                     self, templatePath, baseDir, warnWhenNotFound, expanded
                 )
             else:
-                path, template = self.loadYaml(key, baseDir, warnWhenNotFound)
+                path, template = self.load_yaml(key, baseDir, warnWhenNotFound)
             newBaseDir = os.path.dirname(path)
             if template is None:
                 if warnWhenNotFound:
                     logger.warning(
-                        "document include %s does not exist (base: %s)"
-                        % (templatePath, baseDir)
+                        f"document include {templatePath} does not exist (base: {baseDir})"
                     )
                 return value, template, newBaseDir
             elif isinstance(template, CommentedMap):
-                template.baseDir = newBaseDir
-            _cacheAnchors(self.config._anchorCache, template)
+                template.base_dir = newBaseDir
+            _cache_anchors(self.config._anchorCache, template)
         except:
             raise UnfurlError(
-                "unable to load document include: %s (base: %s)"
-                % (templatePath, baseDir),
-                True,
+                f"unable to load document include: {templatePath} (base: {baseDir})",
                 True,
             )
         self.baseDirs.append(newBaseDir)
